import React from "react";
import { Table } from "react-bootstrap";
import CollapsibleBox from "compendium/components/CollapsibleBox";
import { NrenAndYearDatapoint } from "compendium/Schema";

interface ScrollableTableProps<T> {

    // dataLookup is a map of NRENs -> years -> list of data for that NREN for that year
    dataLookup: Map<string, Map<number, T[]>>;

    // columnInfo is a map of column titles and the lookup key that maps to data for that column
    columnInfo: { [key: string]: string };
}

export function ScrollableTable<T extends NrenAndYearDatapoint>({ dataLookup, columnInfo }: ScrollableTableProps<T>): React.ReactElement {

    if (!dataLookup) {
        return <div className="matrix-border-round"></div>
    }

    const nrenContainers = Array.from(dataLookup.entries()).map(([nren, nrenData]) => {

        return (
            <CollapsibleBox title={nren} key={nren} theme="-table" startCollapsed>
                <div className="scrollable-horizontal">
                    {Array.from(nrenData.entries()).map(([year, yearData]) => {
                        // workaround for setting the background color of the ::before element to the color of the year
                        const style = {
                            '--before-color': `var(--color-of-the-year-muted-${year % 9})`,
                        } as React.CSSProperties

                        return (<div key={year}>
                            <span className={`scrollable-table-year color-of-the-year-${year % 9} bold-caps-16pt pt-3 ps-3`}
                                style={style}>{year}</span>
                            <div className={`colored-table bg-muted-color-of-the-year-${year % 9}`}>
                                <Table>
                                    <thead>
                                        <tr>
                                            {Object.keys(columnInfo).map((title) => (
                                                <th key={title} style={{ position: "relative" }}>
                                                    <span style={style}>{title}</span></th>
                                            ))}
                                        </tr>
                                    </thead>
                                    <tbody>
                                        {yearData.map((datapoint, index) => {
                                            return (
                                                <tr key={index}>
                                                    {Object.entries(columnInfo).map(([title, field]) => {
                                                        const value = datapoint[field];
                                                        return <td key={title}>{value}</td>
                                                    })}
                                                </tr>
                                            );
                                        })}
                                    </tbody>
                                </Table>
                            </div>
                        </div>);
                    })}
                </div>

            </CollapsibleBox>
        );
    });

    return <div className="matrix-border-round">
        {nrenContainers}
    </div>;
}