import { useContext } from 'react';
import { Bar } from 'react-chartjs-2';
import { Chart as ChartJS, CategoryScale, LinearScale, BarElement, Title, Tooltip, Legend } from 'chart.js';

import { Capacity } from "compendium/Schema";
import { createBarChartDataset } from "compendium/helpers/dataconversion";
import DataPage from 'compendium/components/DataPage';
import Filter from "compendium/components/graphing/Filter"
import { Sections } from "compendium/helpers/constants";
import { FilterSelectionContext } from 'compendium/providers/FilterSelectionProvider';
import { useData } from 'compendium/helpers/useData';
import ChartContainer from 'compendium/components/graphing/ChartContainer';
import ChartDataLabels from 'chartjs-plugin-datalabels';
import { getBarChartOptions } from 'compendium/helpers/charthelpers';
import titles from "compendium/titles";

ChartJS.register(
    CategoryScale,
    LinearScale,
    BarElement,
    Title,
    Tooltip,
    Legend
);

function CapacityCoreIPPage() {

    const field = 'typical_backbone_capacity';
    const validityCheck = (data: Capacity) => data[field] != null;
    const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const { data, years, nrens } = useData<Capacity>('/api/capacity', setFilterSelection, validityCheck);

    const selectedData = data.filter(data =>
        filterSelection.selectedYears.includes(data.year) && filterSelection.selectedNrens.includes(data.nren) && validityCheck(data)
    );
    const dataset = createBarChartDataset(selectedData, field, 'Backbone IP Capacity');

    const filterNode = <Filter
        filterOptions={{ availableYears: [...years], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
    />

    const numNrens = Array.from(new Set(selectedData.map(c => c.nren))).length;
    const heightPerBar = 1.5; // every added bar should give this much additional height

    // set a minimum height of 50rem
    const height = Math.max(numNrens * filterSelection.selectedYears.length * heightPerBar + 5, 50);

    const title = titles['capacity-core-ip'];
    const description = `The graph below shows the typical core usable backbone IP capacity of 
    NREN networks, expressed in Gbit/s. It refers to the circuit capacity, not the traffic over 
    the network.`

    const options = getBarChartOptions({

        title: title,
        tooltipUnit: 'Gbit/s',
        unit: 'Gbit/s',
    });

    const filename = "capacity_core_ip";
    return (
        <DataPage title={title}
            description={description}
            category={Sections.Network} filter={filterNode}
            data={selectedData} filename={filename}>

            <ChartContainer>
                <div className="chart-container" style={{ 'height': `${height}rem` }}>
                    <Bar
                        data={dataset}
                        options={options}
                        plugins={[ChartDataLabels]}
                    />
                </div>
            </ChartContainer>
        </DataPage>
    );
}
export default CapacityCoreIPPage;
