import { ReactElement, useContext } from 'react';
import { Row, Col } from "react-bootstrap";
import { Line } from 'react-chartjs-2';
import {
    Chart as ChartJS,
    CategoryScale,
    LinearScale,
    PointElement,
    LineElement,
    Title,
    Tooltip,
    Legend,
} from 'chart.js';

import { TrafficVolume } from "compendium/Schema";
import { createLineChartDataset } from "compendium/helpers/dataconversion";
import DataPage from 'compendium/components/DataPage';
import Filter from "compendium/components/graphing/Filter";
import { Sections } from "compendium/helpers/constants";
import { FilterSelectionContext } from 'compendium/providers/FilterSelectionProvider';
import ChartContainer from "compendium/components/graphing/ChartContainer";
import { useData } from 'compendium/helpers/useData';
import { getLineChartOptions } from 'compendium/helpers/charthelpers';
import titles from "compendium/titles";

ChartJS.register(
    CategoryScale,
    LinearScale,
    PointElement,
    LineElement,
    Title,
    Tooltip,
    Legend
);

function TrafficVolumePage(): ReactElement {

    const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const { data, nrens } = useData<TrafficVolume>('/api/traffic-volume', setFilterSelection);

    const selectedData = data.filter(data =>
        filterSelection.selectedNrens.includes(data.nren) // we only allow filtering nrens for this page
    );

    const fromCustomerTrafficVolumeData = createLineChartDataset(selectedData, 'from_customers');
    const toCustomerTrafficVolumeData = createLineChartDataset(selectedData, 'to_customers');
    const fromExternalTrafficVolumeData = createLineChartDataset(selectedData, 'from_external');
    const toExternalTrafficVolumeData = createLineChartDataset(selectedData, 'to_external');

    const options = getLineChartOptions({
        title: 'Traffic Volume in PB', tooltipUnit: 'PB', valueTransform(value) {
            // show in PB instead
            return value ? value / 1000 : 0;
        },
    });

    const filterNode = <Filter
        filterOptions={{ availableYears: [], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
    />

    return (
        <DataPage title={titles['traffic-volume']}
            description={<span>The four graphs below show the estimates of total annual traffic in PB (1000 TB) to & from NREN customers,
                and to & from external networks. NREN customers are taken to mean sources that are part of the NREN&apos;s connectivity remit,
                while external networks are understood as outside sources including GÉANT, the general/commercial internet, internet
                exchanges, peerings, other NRENs, etc.</span>} category={Sections.Network} filter={filterNode}
            data={selectedData} filename="NREN_traffic_estimates_data">
            <ChartContainer>
                <Row style={{ marginBottom: '30px' }}>
                    <Col>
                        <span style={{ fontSize: '20px', color: 'rgb(85, 96, 156)', fontWeight: 'bold' }}>Traffic from NREN customer</span>
                        <Line data={fromCustomerTrafficVolumeData} options={options} />
                    </Col>
                    <Col>
                        <span style={{ fontSize: '20px', color: 'rgb(221, 100, 57)', fontWeight: 'bold' }}>Traffic to NREN customer</span>
                        <Line data={toCustomerTrafficVolumeData} options={options} />
                    </Col>
                </Row>
                <Row style={{ marginTop: '30px' }}>
                    <Col>
                        <span style={{ fontSize: '20px', color: 'rgb(63, 143, 77)', fontWeight: 'bold' }}>Traffic from external network</span>
                        <Line data={fromExternalTrafficVolumeData} options={options} />
                    </Col>
                    <Col>
                        <span style={{ fontSize: '20px', color: 'rgb(173, 48, 51)', fontWeight: 'bold' }}>Traffic to external network</span>
                        <Line data={toExternalTrafficVolumeData} options={options} />
                    </Col>
                </Row>
            </ChartContainer>
        </DataPage>
    );
}

export default TrafficVolumePage;
