import { useContext } from 'react';
import { Bar } from 'react-chartjs-2';
import { Chart as ChartJS, CategoryScale, LinearScale, BarElement, Title, Tooltip, Legend } from 'chart.js';

import { NrenStaff } from "compendium/Schema";
import { createNRENStaffDatasetAbsolute } from "compendium/helpers/dataconversion";
import DataPage from 'compendium/components/DataPage';
import Filter from "compendium/components/graphing/Filter"
import { Sections } from "compendium/helpers/constants";
import { FilterSelectionContext } from 'compendium/providers/FilterSelectionProvider';
import { useData } from 'compendium/helpers/useData';
import ChartContainer from 'compendium/components/graphing/ChartContainer';
import ChartDataLabels from 'chartjs-plugin-datalabels';
import { getBarChartOptions } from 'compendium/helpers/charthelpers';
import titles from "compendium/titles";

ChartJS.register(
    CategoryScale,
    LinearScale,
    BarElement,
    Title,
    Tooltip,
    Legend
);

function StaffGraphAbsolutePage() {
    const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const { data, years, nrens } = useData<NrenStaff>('/api/staff', setFilterSelection);

    const selectedData = data.filter(data =>
        filterSelection.selectedYears.includes(data.year) && filterSelection.selectedNrens.includes(data.nren)
    );

    const nrenStaffDataset = createNRENStaffDatasetAbsolute(selectedData, filterSelection.selectedYears);

    const filterNode = <Filter
        filterOptions={{ availableYears: [...years], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
    />

    const numNrens = Array.from(new Set(selectedData.map(c => c.nren))).length;
    const heightPerBar = 1.5; // every added bar should give this much additional height

    // set a minimum height of 50rem
    const height = Math.max(numNrens * filterSelection.selectedYears.length * heightPerBar + 5, 50);

    const title = titles["employee-count"];
    const description = `The graph shows the total number of employees (in FTEs) at each NREN. When filling in the survey, NRENs are asked about the number of staff engaged (whether permanent or subcontracted) in NREN activities. Please note that diversity within the NREN community means that there is not one single definition of what constitutes "NREN activities". Therefore due to differences in how their organisations are arranged, and the relationship, in some cases, with parent organisations, there can be inconsistencies in how NRENs approach this question.`

    const options = getBarChartOptions({ tooltipPrefix: "FTEs", title: "Full-Time Equivalents" });
    const filename = "number_of_nren_employees";
    return (
        <DataPage title={title}
            description={description}
            category={Sections.Organisation} filter={filterNode}
            data={selectedData} filename={filename}>

            <ChartContainer>
                <div className="chart-container" style={{ 'height': `${height}rem` }}>
                    <Bar
                        data={nrenStaffDataset}
                        options={options}
                        plugins={[ChartDataLabels]}
                    />
                </div>
            </ChartContainer>
        </DataPage>
    );
}
export default StaffGraphAbsolutePage;
