"""
Utilities used around the codebase.
"""
import re

from flask import request
from flask_login import current_user  # type: ignore
from typing import Type, Sequence, TypeVar, Dict, Any
from itertools import chain

from compendium_v2.db import db
from compendium_v2.db.presentation_models import (
    NREN, NRENExclusion, PreviewYear, PresentationModel, ExternalConnections)
from sqlalchemy import select

T = TypeVar('T', bound=PresentationModel)


def get_data(table_class: Type[T]) -> Sequence[T]:
    if table_class == NREN:
        data = [extract_model_data(n) for n in db.session.scalars(select(NREN).order_by(NREN.name.asc()))]
        return list(chain(*data))
    select_statement = select(table_class).join(NREN).order_by(NREN.name.asc(), table_class.year.asc()).where(
        ~table_class.nren_id.in_(select(NRENExclusion.nren_id).where(NRENExclusion.year == table_class.year))
    )

    can_preview = (not current_user.is_anonymous) and current_user.can_preview
    preview = can_preview and request.args.get('preview') is not None
    if not preview:
        select_statement = select_statement.where(table_class.year.not_in(select(PreviewYear.year)))

    data = [extract_model_data(m) for m in db.session.scalars(select_statement)]
    return list(chain(*data))


def extract_model_data(entry_model, download=False):
    if isinstance(entry_model, NREN):
        # if we fetch NREN specifically, send the ID as it's needed for management
        yield entry_model.to_dict_with_id()
        return

    if download or isinstance(entry_model, ExternalConnections):
        # if we're downloading data, we need to flatten the data structure
        # always flatten ExternalConnections
        try:
            iter(entry_model)
            yield from entry_model.to_list(download=download)
        except TypeError:
            yield entry_model.to_dict(download=download)
        return

    yield entry_model.to_dict()


def get_survey_question_dependency_graph(survey: Dict[str, Any]):
    """
    This function takes a survey in JSON format, walks through the questions and builds a dependency graph of which
    questions depend on which other questions in various ways.
    Mostly centered around visibleIf logic, to make sure that when moving data from one survey to the next, only
    keys that are actually used are moved.

    E.g. if a question is not answered, but a hidden question depends on it, the data for the hidden question
    should not be moved to the next survey response "prefilled" automatically.
    """

    pages = survey.get('pages', [])
    if not pages:
        return {}

    dependencies: Dict[str, Any] = {}

    def _extract_dependencies(visibleIf: str):
        """
        Extract the key that this question depends on. There's a few cases covered:

        It's usually structured like:
        {network_automation} = 'stringvalue'
        {network_automation} = ['stringvalue in array']
        {network_automation} = 'yes' or {network_automation} = 'planned'

        This function extracts the key that the question depends on, as well as the required value.
        In case of multiple tests, a tuple is returned with the key and the accepted values.

        """

        if not visibleIf:
            return None

        assert isinstance(visibleIf, str), f"visibleIf should be a string, got {visibleIf}"

        is_or = False
        if ' or ' in visibleIf:
            is_or = True
            groups = visibleIf.split(' or ')
        elif ' and ' in visibleIf:
            groups = visibleIf.split(' and ')
        else:
            groups = []

        if len(groups) > 1:
            groups = [group.strip() for group in groups]
            operation = 'or' if is_or else 'and'
            return operation, tuple(_extract_dependencies(group) for group in groups)

        match = re.match(r'{(.+?)}\s*=\s*(.+)', visibleIf)
        if not match:
            raise ValueError(f"Could not parse visibleIf: {visibleIf}")

        key, value = match.groups()
        key = key.strip().replace('row.', '')
        value = value.strip()

        if value.startswith('[') and value.endswith(']'):
            # "['yes']" -> ['yes'] as a python list
            value = [value[1:-1].replace("'", "").replace('"', "")]
        else:
            # "'yes'" -> "yes"
            value = value.replace("'", "").replace('"', "")

        return key, value

    def _add_element_dependencies(question):
        name = question.get('name')

        if not name:
            raise ValueError(f"Question {question} does not have a name, cannot determine dependencies")

        deps = dependencies.get(name, [])

        visible_if = question.get('visibleIf')
        if not visible_if:
            return

        dependent_key = _extract_dependencies(visible_if)
        if dependent_key:
            deps.append(dependent_key)

        if deps:
            dependencies[name] = deps

    def _process_element(element):
        if element.get('type') in ['comment', 'html']:
            return
        if element.get('type') == 'panel':
            panel_elements = element.get('elements', [])
            if not panel_elements:
                return

            for panel_element in panel_elements:
                _process_element(panel_element)
        if element.get('type') == 'matrixdropdown':
            columns = element.get('columns', [])
            if not columns:
                return

            for column in columns:
                if not column.get('visibleIf'):
                    continue
                column['name'] = f"{element.get('name')}[..]{column.get('name')}"
                _add_element_dependencies(column)
        else:
            _add_element_dependencies(element)

    for page in pages:
        elements = page.get('elements', [])
        if not elements:
            continue

        for element in elements:
            _process_element(element)

    return dependencies


def clean_answers_with_dependencies(data: Dict[str, Any], survey: Dict[str, Any]):
    """
    Given a survey and a set of answers, clean the answers based on the dependencies in the survey.
    This is used to clean up the answers before moving them to the next survey, to make sure that only the
    necessary answers are moved.
    """

    if not survey:
        raise ValueError("Provided survey is empty, cannot clean answers")

    dependencies = get_survey_question_dependency_graph(survey)

    cleaned_data = data.copy()

    for key, value in dependencies.items():

        if '[..]' in key:
            # matrixdropdown
            matrix_key, column_key = key.split('[..]')
            if matrix_key not in data:
                continue

            matrix_data = data.get(matrix_key)
            if not matrix_data:
                continue

            for column, row in list(matrix_data.items()):
                if column_key not in row:
                    continue

                is_valid = column_key in row
                for expected_key, values in value:
                    if not (expected_key in row and row[expected_key] == values):
                        is_valid = False
                        break

                if is_valid:
                    cleaned_data.setdefault(matrix_key, {})[column][column_key] = row[column_key]
                else:
                    del cleaned_data[matrix_key][column][column_key]

            continue
        if key not in data:
            continue

        data_value = data.get(key)
        # check that the data_value satisfies the dependency

        if not isinstance(value, list):
            raise ValueError(f"Expected dependencies to be a list, got {value}")

        # multiple dependencies, all must be satisfied
        is_valid = True
        for expected_key, values in value:
            actual_value = data.get(expected_key)
            if isinstance(actual_value, str):
                actual_value = actual_value.lower()  # survey is case insensitive
            if isinstance(values, str):
                values = values.lower()

            if expected_key in ['or', 'and']:
                # special handling for when a question is shown if one or more conditions are met

                if expected_key == 'or':
                    or_valid = False
                    # if any of the conditions are met, the question is shown
                    for _or_expected_key, _or_values in values:
                        _or_data = data.get(_or_expected_key)
                        if isinstance(_or_values, str):
                            _or_values = _or_values.lower()
                        if isinstance(_or_data, str):
                            _or_data = _or_data.lower()

                        if _or_expected_key in data and _or_data == _or_values:
                            or_valid = True
                            break
                    if not or_valid:
                        is_valid = False
                        break
                elif expected_key == 'and':
                    and_valid = True
                    # all conditions must be met for the question to be shown
                    for _and_expected_key, _and_values in values:
                        _and_data = data.get(_and_expected_key)
                        if isinstance(_and_values, str):
                            _and_values = _and_values.lower()
                        if isinstance(_and_data, str):
                            _and_data = _and_data.lower()

                        if not (_and_expected_key in data and _and_data == _and_values):
                            and_valid = False
                            break
                    if not and_valid:
                        is_valid = False
                        break

            elif not (expected_key in data and actual_value == values):
                is_valid = False
                break

        if is_valid:
            cleaned_data[key] = data_value
        else:
            del cleaned_data[key]

    return cleaned_data
