"""
survey_publisher_legacy_db
============================

This module loads the survey data from 2022 from the old survey database and returns the data
ready to be used to generate new-style surveys.

"""
import logging
import enum
import json
import html
import itertools
import os
import csv

from decimal import Decimal
from sqlalchemy import text
from collections import defaultdict

from compendium_v2.conversion.mapping import CHARGING_LEVELS, CONNECTION, INTERCONNECTION, SERVICE_USER_TYPE_TO_CODE
from compendium_v2.db.presentation_model_enums import CommercialCharges, CommercialConnectivityCoverage, \
    ConnectionMethod, FeeType, ServiceCategory, UserCategory, YesNoPlanned
from compendium_v2.environment import setup_logging
from compendium_v2.publishers.helpers import extract_urls, valid_url
from compendium_v2.survey_db import model as survey_model
from compendium_v2.db import db
from compendium_v2.publishers import helpers
from compendium_v2.publishers.helpers import map_nren
from compendium_v2.conversion import mapping

setup_logging()

logger = logging.getLogger('survey-publisher-old-db-2022')

BUDGET_QUERY = """
SELECT DISTINCT ON (n.id, a.question_id)
    n.abbreviation AS nren,
    a.value AS budget
FROM answers a
JOIN nrens n ON a.nren_id = n.id
JOIN questions q ON a.question_id = q.id
JOIN sections s ON q.section_id = s.id
JOIN compendia c ON s.compendium_id = c.id
WHERE
    a.question_id = 16402
AND c.year = 2022
ORDER BY n.id, a.question_id, a.updated_at DESC
"""

QUESTION_TEMPLATE_QUERY = """
SELECT DISTINCT ON (n.id, a.question_id)
    n.abbreviation AS nren,
    a.value AS value
FROM answers a
JOIN nrens n ON a.nren_id = n.id
JOIN questions q ON a.question_id = q.id
JOIN sections s ON q.section_id = s.id
JOIN compendia c ON s.compendium_id = c.id
WHERE
    a.question_id = {}
    AND c.year = {}
    AND a.value NOT IN ('"NA"', '"N/A"', '[""]', '["-"]', '["/"]')
ORDER BY n.id, a.question_id, a.updated_at DESC
"""

RECURSIVE_QUERY = """
    WITH RECURSIVE parent_questions AS (
        -- Base case
        SELECT q.id, q.equivalent_question_id, c.year, q.title
        FROM questions q
                 JOIN sections s ON q.section_id = s.id
                 JOIN compendia c ON s.compendium_id = c.id
        WHERE q.id = {}
        UNION ALL
        -- Recursive case
        SELECT q.id, q.equivalent_question_id, c.year, q.title
        FROM questions q
                 INNER JOIN parent_questions pq ON q.id = pq.equivalent_question_id
                 JOIN sections s ON q.section_id = s.id
                 JOIN compendia c ON s.compendium_id = c.id)
    SELECT DISTINCT ON (n.id, answers.question_id) answers.id,
                                                   UPPER(n.abbreviation) AS nren,
                                                   parent_questions.year,
                                                   answers.value         as answer
    FROM answers
             JOIN parent_questions ON answers.question_id = parent_questions.id
             JOIN nrens n on answers.nren_id = n.id
    WHERE UPPER(answers.value) NOT IN ('"NA"', '"N/A"', '[""]', '["-"]', '["/"]', '/', '["NA"]', '""', '[]', '[n/a]')
    ORDER BY n.id, answers.question_id, answers.updated_at DESC;
"""

SECURITY_CONTROLS_CSV = os.path.join(os.path.dirname(__file__), "../../resources/security_controls_parsed.csv")


class FundingSource(enum.Enum):
    CLIENT_INSTITUTIONS = 16405
    EUROPEAN_FUNDING = 16406
    COMMERCIAL = 16407
    OTHER = 16408
    GOV_PUBLIC_BODIES = 16409


class StaffQuestion(enum.Enum):
    """
    Answers are numbers expressed in FTEs (full time equivalents)
    """
    PERMANENT_FTE = 16414
    SUBCONTRACTED_FTE = 16413
    TECHNICAL_FTE = 16416
    NON_TECHNICAL_FTE = 16417


class OrgQuestion(enum.Enum):
    """
    Answers are strings
    """
    PARENT_ORG_NAME = 16419

    SUB_ORGS_1_NAME = 16422
    SUB_ORGS_1_CHOICE = 16449
    SUB_ORGS_1_ROLE = 16426

    SUB_ORGS_2_NAME = 16429
    SUB_ORGS_2_CHOICE = 16448
    SUB_ORGS_2_ROLE = 16434

    SUB_ORGS_3_NAME = 16430
    SUB_ORGS_3_CHOICE = 16446
    SUB_ORGS_3_ROLE = 16435

    SUB_ORGS_4_NAME = 16432
    SUB_ORGS_4_CHOICE = 16451
    SUB_ORGS_4_ROLE = 16438

    SUB_ORGS_5_NAME = 16433
    SUB_ORGS_5_CHOICE = 16450
    SUB_ORGS_5_ROLE = 16439


class ECQuestion(enum.Enum):
    EC_PROJECT = 16453


class ChargingStructure(enum.Enum):
    """
    Answers are strings
    """
    charging_structure = 16410


def query_budget():
    return db.session.execute(text(BUDGET_QUERY), bind_arguments={'bind': db.engines[survey_model.SURVEY_DB_BIND]})


def recursive_query(question_id_2022):
    assert question_id_2022
    query = RECURSIVE_QUERY.format(question_id_2022)

    data = db.session.execute(text(query), bind_arguments={'bind': db.engines[survey_model.SURVEY_DB_BIND]})
    for q_id, nren, year, answer in data:
        yield (q_id, map_nren(nren), year, answer)


def query_funding_sources():
    for source in FundingSource:
        query = QUESTION_TEMPLATE_QUERY.format(source.value, 2022)
        yield source, db.session.execute(text(query), bind_arguments={'bind': db.engines[survey_model.SURVEY_DB_BIND]})


def query_question(question: enum.Enum):
    return query_question_id(question.value)


def query_question_id(question_id: int, year: int = 2022):
    query = QUESTION_TEMPLATE_QUERY.format(question_id, year)
    bind = {'bind': db.engines[survey_model.SURVEY_DB_BIND]}
    for abbrev, answer in db.session.execute(text(query), bind_arguments=bind):
        yield map_nren(abbrev), answer


def budget(nren_dict):
    rows = query_budget()
    for row in rows:
        nren_name = row[0].upper()
        _budget = row[1]
        try:
            budget = float(_budget.replace('"', '').replace(',', ''))
        except ValueError:
            continue

        if nren_name not in nren_dict:
            continue

        yield ('budget', nren_dict[nren_name], nren_dict[nren_name].id, 2022, budget)


def funding_sources(nren_dict):
    sourcedata = {}
    for source, data in query_funding_sources():
        for row in data:
            nren_name = row[0].upper()
            _value = row[1]
            try:
                value = float(_value.replace('"', '').replace(',', ''))
            except ValueError:
                value = 0

            nren_info = sourcedata.setdefault(
                nren_name,
                {source_type: 0 for source_type in FundingSource}
            )
            nren_info[source] = value

    for nren_name, nren_info in sourcedata.items():
        if nren_name not in nren_dict:
            continue

        nren = nren_dict[nren_name]

        if nren.name.lower() == 'heanet':
            nren_info[FundingSource.OTHER] = 77.78
            nren_info[FundingSource.COMMERCIAL] = 0

        data = {
            'client_institutions': round(nren_info[FundingSource.CLIENT_INSTITUTIONS], 2),
            'european_funding': round(nren_info[FundingSource.EUROPEAN_FUNDING], 2),
            'commercial': round(nren_info[FundingSource.COMMERCIAL], 2),
            'other': round(nren_info[FundingSource.OTHER], 2),
            'gov_public_bodies': round(nren_info[FundingSource.GOV_PUBLIC_BODIES], 2),
        }

        yield ('income_sources', nren, nren.id, 2022, data)


def charging_structure(nren_dict):
    rows = query_question(ChargingStructure.charging_structure)
    for row in rows:
        nren_name = row[0].upper()
        value = row[1].replace('"', '').strip()

        if nren_name not in nren_dict:
            continue

        if "do not charge" in value:
            charging_structure = FeeType.no_charge
        elif "combination" in value:
            charging_structure = FeeType.combination
        elif "flat" in value:
            charging_structure = FeeType.flat_fee
        elif "usage-based" in value:
            charging_structure = FeeType.usage_based_fee
        elif "Other" in value:
            charging_structure = FeeType.other
        else:
            charging_structure = None

        if charging_structure:
            charging_structure = charging_structure.name

        yield ('charging_mechanism', nren_dict[nren_name], nren_dict[nren_name].id, 2022, charging_structure)


def staff_data(nren_dict):
    data = {}
    for question in StaffQuestion:
        rows = query_question(question)
        for row in rows:
            nren_name = row[0].upper()
            _value = row[1]
            try:
                value = float(_value.replace('"', '').replace(',', ''))
            except ValueError:
                value = 0

            if nren_name not in nren_dict:
                continue

            # initialize on first use, so we don't add data for nrens with no answers
            data.setdefault(nren_name, {question: 0 for question in StaffQuestion})[question] = value

    for nren_name, nren_info in data.items():
        if sum([nren_info[question] for question in StaffQuestion]) == 0:
            continue

        staff_type_data = {
            'permanent_fte': nren_info[StaffQuestion.PERMANENT_FTE],
            'subcontracted_fte': nren_info[StaffQuestion.SUBCONTRACTED_FTE],
        }

        staff_roles_data = {
            'technical_fte': nren_info[StaffQuestion.TECHNICAL_FTE],
            'non_technical_fte': nren_info[StaffQuestion.NON_TECHNICAL_FTE],
        }

        yield ('staff_employment_type', nren_dict[nren_name], nren_dict[nren_name].id, 2022, staff_type_data)

        yield ('staff_roles', nren_dict[nren_name], nren_dict[nren_name].id, 2022, staff_roles_data)


def nren_parent_org(nren_dict):
    # clean up the data a bit by removing some strings
    strings_to_replace = [
        'We are affiliated to '
    ]

    rows = list(query_question(OrgQuestion.PARENT_ORG_NAME))

    for row in rows:
        nren_name = row[0].upper()
        value = str(row[1]).replace('"', '')

        if not value:
            continue

        for string in strings_to_replace:
            value = value.replace(string, '')

        if nren_name not in nren_dict:
            continue

        yield ('parent_organization', nren_dict[nren_name], nren_dict[nren_name].id, 2022, 'Yes')
        yield ('parent_organization_name', nren_dict[nren_name], nren_dict[nren_name].id, 2022, value)


def nren_sub_org(nren_dict):
    suborg_questions = [
        (OrgQuestion.SUB_ORGS_1_NAME, OrgQuestion.SUB_ORGS_1_CHOICE, OrgQuestion.SUB_ORGS_1_ROLE),
        (OrgQuestion.SUB_ORGS_2_NAME, OrgQuestion.SUB_ORGS_2_CHOICE, OrgQuestion.SUB_ORGS_2_ROLE),
        (OrgQuestion.SUB_ORGS_3_NAME, OrgQuestion.SUB_ORGS_3_CHOICE, OrgQuestion.SUB_ORGS_3_ROLE),
        (OrgQuestion.SUB_ORGS_4_NAME, OrgQuestion.SUB_ORGS_4_CHOICE, OrgQuestion.SUB_ORGS_4_ROLE),
        (OrgQuestion.SUB_ORGS_5_NAME, OrgQuestion.SUB_ORGS_5_CHOICE, OrgQuestion.SUB_ORGS_5_ROLE)
    ]
    lookup = defaultdict(list)

    for name, choice, role in suborg_questions:
        _name_rows = query_question(name)
        _choice_rows = query_question(choice)
        _role_rows = list(query_question(role))
        for _name, _choice in zip(_name_rows, _choice_rows):
            nren_name = _name[0].upper()
            suborg_name = _name[1].replace('"', '').strip()
            role_choice = _choice[1].replace('"', '').strip()

            if nren_name not in nren_dict:
                continue

            if role_choice.lower() == 'other':
                for _role in _role_rows:
                    if _role[0] == _name[0]:
                        role = _role[1].replace('"', '').strip()
                        break
            else:
                role = role_choice

            if not role:
                continue

            lookup[nren_name].append((suborg_name, role))

    for nren_name, suborgs in lookup.items():
        if suborgs:
            yield ('suborganizations', nren_dict[nren_name], nren_dict[nren_name].id, 2022, 'Yes')

        result = []
        for suborg_name, role in suborgs:
            result.append({
                'suborganization_name': suborg_name,
                'suborganization_role': role
            })

        yield ('suborganization_details', nren_dict[nren_name], nren_dict[nren_name].id, 2022, result)


def ec_projects(nren_dict):
    rows = query_question(ECQuestion.EC_PROJECT)
    for row in rows:
        nren_name = row[0].upper()

        if nren_name not in nren_dict:
            continue

        try:
            value = json.loads(row[1])
        except json.decoder.JSONDecodeError:
            continue

        has_values = any(value)

        if has_values:
            yield ('ec_projects', nren_dict[nren_name], nren_dict[nren_name].id, 2022, "Yes")

        result = []

        for val in value:
            if not val:
                continue

            # strip html entities/NBSP from val
            val = html.unescape(val).replace('\xa0', ' ')

            # some answers include contract numbers, which we don't want here
            val = val.split('(contract n')[0]
            result.append(str(val).strip())
        projects = [{'ec_project_name': r} for r in result]
        yield ('ec_project_names', nren_dict[nren_name], nren_dict[nren_name].id, 2022, projects)


def policies(nren_dict):
    """
    Answers are strings that should be urls, but sometimes there's other stuff
    like email addresses or random text
    """

    policy_questions = {
        'strategy': {2022: 16469, 2021: 16064, 2020: 15720, 2019: 15305, 2018: 14910},
        'environment': {2022: 16471, 2021: 16066, 2020: 15722, 2019: 15307, 2018: 14912},
        'equality': {2022: 16473, 2021: 16378},
        'connectivity': {2022: 16475, 2021: 16068, 2020: 15724, 2019: 15309, 2018: 14914},
        'acceptable_use': {2022: 16477, 2021: 16070, 2020: 15726, 2019: 15311, 2018: 14916},
        'privacy': {2022: 16479, 2021: 16072, 2020: 15728, 2019: 15575},
        'data_protection': {2022: 16481, 2021: 16074, 2020: 15730, 2019: 15577},
        'gender': {2022: 16761}
    }

    data = {}
    for year in [2018, 2019, 2020, 2021, 2022]:
        policy_questions_year = {key: years[year] for key, years in policy_questions.items() if year in years}
        for question_key, question_id in policy_questions_year.items():
            rows = query_question_id(question_id, year)
            for row in rows:
                nren_name = row[0].upper()
                _value = row[1]

                if nren_name not in nren_dict:
                    continue

                value = _value.split()[0].strip('"')

                if not value:
                    continue

                if value.upper() == 'N.A.' or ('.' not in value and '@' not in value):
                    # this test is a bit silly but does seem to filter out all the nonsense responses
                    continue

                if _value not in [f'"{value}"', value]:
                    pass

                # initialize on first use, so we don't add data for nrens with no answers
                data.setdefault((nren_name, year), {q: '' for q in policy_questions.keys()})
                data[(nren_name, year)][question_key] = value

    for (nren_name, year), nren_info in data.items():

        strategy = nren_info['strategy']
        if strategy and not valid_url(strategy):
            strategy = ''
        environment = nren_info['environment']
        if environment and not valid_url(environment):
            environment = ''
        equality = nren_info['equality']
        if equality and not valid_url(equality):
            equality = ''
        connectivity = nren_info['connectivity']
        if connectivity and not valid_url(connectivity):
            connectivity = ''
        acceptable_use = nren_info['acceptable_use']
        if acceptable_use and not valid_url(acceptable_use):
            acceptable_use = ''
        privacy = nren_info['privacy']
        if privacy and not valid_url(privacy):
            privacy = ''
        data_protection = nren_info['data_protection']
        if data_protection and not valid_url(data_protection):
            data_protection = ''
        gender_equality = nren_info['gender']
        if gender_equality and not valid_url(gender_equality):
            gender_equality = ''

        all_policies = [strategy, environment, equality, connectivity,
                        acceptable_use, privacy, data_protection, gender_equality]
        if not any(all_policies):
            continue

        if strategy:
            yield ('corporate_strategy', nren_dict[nren_name], nren_dict[nren_name].id, year, 'Yes')
            yield ('corporate_strategy_url', nren_dict[nren_name], nren_dict[nren_name].id, year, strategy)

        policies = {}
        if environment:
            policies['environmental_policy'] = {
                'available': ['yes'],
                'url': environment
            }
        if equality:
            policies['equal_opportunity_policy'] = {
                'available': ['yes'],
                'url': equality
            }
        if connectivity:
            policies['connectivity_policy'] = {
                'available': ['yes'],
                'url': connectivity
            }
        if acceptable_use:
            policies['acceptable_use_policy'] = {
                'available': ['yes'],
                'url': acceptable_use
            }
        if privacy:
            policies['privacy_notice'] = {
                'available': ['yes'],
                'url': privacy
            }
        if data_protection:
            policies['data_protection_contact'] = {
                'available': ['yes'],
                'url': data_protection
            }
        if gender_equality:
            policies['gender_equality_policy'] = {
                'available': ['yes'],
                'url': gender_equality
            }

        yield ('policies', nren_dict[nren_name], nren_dict[nren_name].id, year, policies)


def central_procurement(nren_dict):

    rows = recursive_query(16482)
    amounts = recursive_query(16483)
    amounts = {(nren_name, year): float(answer.strip('"')) for answer_id, nren_name, year, answer in amounts}

    for answer_id, nren_name, year, answer in rows:
        if nren_name not in nren_dict:
            continue

        answer = 'Yes' if answer == '"Yes"' else 'No'
        yield ('central_software_procurement', nren_dict[nren_name], nren_dict[nren_name].id, year, answer)

        amount = amounts.get((nren_name, year))
        if answer == 'Yes':
            yield ('central_procurement_amount', nren_dict[nren_name], nren_dict[nren_name].id, year, amount)


def service_management(nren_dict):

    framework = recursive_query(16484)
    framework = {(nren_name, year): answer == '"Yes"' for answer_id, nren_name, year, answer in framework}
    targets = recursive_query(16485)
    targets = {(nren_name, year): answer == '"Yes"' for answer_id, nren_name, year, answer in targets}

    for nren_name, year in framework.keys() | targets.keys():
        if nren_name not in nren_dict:
            continue

        nren = nren_dict[nren_name]
        framework_val = framework.get((nren_name, year))
        target_value = targets.get((nren_name, year))
        yield ('formal_service_management_framework', nren, nren.id, year, 'Yes' if framework_val else 'No')
        yield ('service_level_targets', nren, nren.id, year, 'Yes' if target_value else 'No')


def service_user_types(nren_dict):

    categories = [
        (ServiceCategory.identity, 16488),
        (ServiceCategory.network_services, 16489),
        (ServiceCategory.collaboration, 16490),
        (ServiceCategory.security, 16491),
        (ServiceCategory.isp_support, 16492),
        (ServiceCategory.storage_and_hosting, 16493),
        (ServiceCategory.multimedia, 16494),
        (ServiceCategory.professional_services, 16495)
    ]
    by_nren = defaultdict(lambda: defaultdict(dict))
    for service_category, question_id in categories:
        rows = recursive_query(question_id)
        for answer_id, nren_name, year, answer in rows:
            if nren_name not in nren_dict:
                continue

            service_matrix = by_nren.setdefault(nren_name, {}).setdefault(year, {})

            for user_cat_db in json.loads(answer):
                user_cat = UserCategory[SERVICE_USER_TYPE_TO_CODE[user_cat_db]]
                category_data = service_matrix.setdefault(user_cat.name, {})
                service_types = category_data.setdefault('service_types', [])

                service_types.append(service_category.name)

    for nren_name, yeardata in by_nren.items():
        for year, service_matrix in yeardata.items():
            if not service_matrix:
                continue
            yield ('service_matrix', nren_dict[nren_name], nren_dict[nren_name].id, year, service_matrix)


def standards(nren_dict):

    audits = recursive_query(16499)
    audits = {(nren_name, year): answer == '"Yes"' for answer_id, nren_name, year, answer in audits}
    audit_specifics = recursive_query(16500)
    audit_specifics = {(nren_name, year): answer.strip('"') for answer_id, nren_name, year, answer in audit_specifics}
    bcp = recursive_query(16501)
    bcp = {(nren_name, year): answer == '"Yes"' for answer_id, nren_name, year, answer in bcp}
    bcp_specifics = recursive_query(16502)
    bcp_specifics = {(nren_name, year): answer.strip('"') for answer_id, nren_name, year, answer in bcp_specifics}
    cmp = recursive_query(16762)
    cmp = {(nren_name, year): answer == '"Yes"' for answer_id, nren_name, year, answer in cmp}

    for nren_name, year in audits.keys() | audit_specifics.keys() | bcp.keys() | bcp_specifics.keys() | cmp.keys():
        if nren_name not in nren_dict:
            continue

        nren = nren_dict[nren_name]

        audit = audits.get((nren_name, year))
        a_specifics = audit_specifics.get((nren_name, year))
        bcp_val = bcp.get((nren_name, year))
        bcp_s_val = bcp_specifics.get((nren_name, year))
        cmp_val = cmp.get((nren_name, year))

        if all([v is None or v == '' for v in [audit, a_specifics, bcp_val, bcp_s_val, cmp_val]]):
            continue

        if audit is not None:
            audit = 'Yes' if audit else 'No'
        yield ('audits', nren, nren.id, year, audit)

        yield ('audit_specifics', nren, nren.id, year, a_specifics or '')

        if bcp_val is not None:
            bcp_val = 'Yes' if bcp_val else 'No'
        yield ('business_continuity_plans', nren, nren.id, year, bcp_val)

        yield ('business_continuity_plans_specifics', nren, nren.id, year, bcp_s_val or '')

        if cmp_val is not None:
            cmp_val = 'Yes' if cmp_val else 'No'
        yield ('crisis_management_procedure', nren, nren.id, year, cmp_val)


def crisis_exercises(nren_dict):

    rows = recursive_query(16763)

    crisis_exercises_map = {
        "geant_workshops": "We participate in GEANT Crisis workshops such as CLAW",
        "national_excercises": "We participated in National crisis exercises ",
        "tabletop_exercises": "We run our own tabletop exercises",
        "simulation_excercises": "We run our own simulation exercises",
        "other_excercises": "We have done/participated in other exercises or trainings",
        "real_crisis": "We had a real crisis",
        "internal_security_programme": "We run an internal security awareness programme",
        "none": "No, we have not done any crisis exercises or trainings",
    }
    _reversed_map = {v: k for k, v in crisis_exercises_map.items()}

    for answer_id, nren_name, year, answer in rows:
        if nren_name not in nren_dict:
            continue

        descriptions = list(filter(lambda d: bool(d), [_reversed_map.get(desc) for desc in json.loads(answer)]))

        yield ('crisis_exercises', nren_dict[nren_name], nren_dict[nren_name].id, year, descriptions)


def security_controls(nren_dict):

    controls_map = {
        "anti_virus": "Anti Virus",
        "anti_spam": "Anti-Spam",
        "firewall": "Firewall",
        "ddos_mitigation": "DDoS mitigation",
        "monitoring": "Network monitoring",
        "ips_ids": "IPS/IDS",
        "acl": "ACL",
        "segmentation": "Network segmentation",
        "integrity_checking": "Integrity checking"
    }

    reversed_map = {v: k for k, v in controls_map.items()}

    sc = recursive_query(16503)
    sc = {(nren_name, year): json.loads(answer) for answer_id, nren_name, year, answer in sc}
    sc_other = recursive_query(16504)
    sc_other = {(nren_name, year): json.loads(answer) for answer_id, nren_name, year, answer in sc_other}
    for key, value in sc_other.items():
        if not isinstance(value, list):
            sc_other[key] = [value]

    for nren_name, year in sc.keys() | sc_other.keys():
        if year < 2021:  # prior to 2022, the mapping is different, use a different data source
            continue

        if nren_name not in nren_dict:
            continue

        nren = nren_dict[nren_name]

        full_list = sc.get((nren_name, year), [])
        other_entries = [e.strip() for e in sc_other.get((nren_name, year), [])
                         if e.strip() and e.lower() not in ["n/a", "-"]]
        other_entry = ", ".join(other_entries)
        if other_entry:
            full_list.append(other_entry)
            if "Other" in full_list:
                full_list.remove("Other")

        full_list = list(filter(lambda d: bool(d), [reversed_map.get(control) for control in full_list]))

        if other_entry:
            full_list.append('other')

        yield ('security_controls', nren, nren.id, year, full_list)
        if other_entry:
            yield ('security_controls-Comment', nren, nren.id, year, other_entry)

    # pop on data manually fixed from free-text fields for <2021
    with open(SECURITY_CONTROLS_CSV, 'r') as f:
        reader = csv.DictReader(f)
        for row in reader:
            nren_name = row['nren'].upper()
            year = int(row['year'])
            if nren_name not in nren_dict:
                continue

            nren = nren_dict[nren_name]

            controls = {**row}
            del controls['nren']
            del controls['year']

            other = ''
            if controls.get('other'):
                other = controls['other'].strip()
                del controls['other']

            controls_for_year = [k for k, v in controls.items() if v]

            yield ('security_controls', nren, nren.id, year, controls_for_year)
            if other:
                yield ('security_controls-Comment', nren, nren.id, year, other)


def institutions_urls(nren_dict):

    rows = recursive_query(16507)

    for row in rows:
        answer_id, nren_name, year, answer = row
        if nren_name not in nren_dict:
            continue

        urls = extract_urls(text=answer)

        if not urls:
            continue

        valid_urls = []

        for url in urls:
            if not valid_url(url):
                continue
            valid_urls.append(url)

        if not valid_urls:
            continue

        connected_sites = []
        for url in valid_urls:
            connected_sites.append({'connected_sites_url': url})

        yield ('connected_sites_lists', nren_dict[nren_name], nren_dict[nren_name].id, year, connected_sites)


def commercial_connectivity(nren_dict):
    simple_connection = {
        key.replace(" ", "").replace("-", "").replace("/", "").lower(): value for key, value in CONNECTION.items()
    }

    def get_coverage(db_string):
        cleaned_str = db_string.strip('"').replace(" ", "").replace("-", "").replace("/", "").lower()
        key = simple_connection[cleaned_str]
        return CommercialConnectivityCoverage[key]

    sp = recursive_query(16646)
    sp = {(nren_name, year): get_coverage(answer) for answer_id, nren_name, year, answer in sp}
    collab = recursive_query(16647)
    collab = {(nren_name, year): get_coverage(answer) for answer_id, nren_name, year, answer in collab}
    r_e = recursive_query(16648)
    r_e = {(nren_name, year): get_coverage(answer) for answer_id, nren_name, year, answer in r_e}
    general = recursive_query(16649)
    general = {(nren_name, year): get_coverage(answer) for answer_id, nren_name, year, answer in general}
    spin_off = recursive_query(16650)
    spin_off = {(nren_name, year): get_coverage(answer) for answer_id, nren_name, year, answer in spin_off}

    for nren_name, year in sp.keys() | collab.keys() | r_e.keys() | general.keys() | spin_off.keys():
        if nren_name not in nren_dict:
            continue

        commercial_orgs = {}

        commercial_r_e = r_e.get((nren_name, year))
        commercial_general = general.get((nren_name, year))
        commercial_collaboration = collab.get((nren_name, year))
        commercial_service_provider = sp.get((nren_name, year))
        university_spin_off = spin_off.get((nren_name, year))

        if commercial_r_e:
            commercial_orgs['commercial_r_e'] = {
                'connection': commercial_r_e.name
            }
        if commercial_general:
            commercial_orgs['commercial_general'] = {
                'connection': commercial_general.name
            }
        if commercial_collaboration:
            commercial_orgs['commercial_collaboration'] = {
                'connection': commercial_collaboration.name
            }
        if commercial_service_provider:
            commercial_orgs['commercial_service_provider'] = {
                'connection': commercial_service_provider.name
            }
        if university_spin_off:
            commercial_orgs['university_spin_off'] = {
                'connection': university_spin_off.name
            }

        yield ('commercial_organizations', nren_dict[nren_name], nren_dict[nren_name].id, year, commercial_orgs)


def commercial_charging_level(nren_dict):

    simple_charging = {
        key.replace(" ", "").replace("-", "").replace("/", "").lower(): value for key, value in CHARGING_LEVELS.items()
    }
    simple_charging["nochargesapplied"] = "no_charges_if_r_e_requested"
    simple_charging['nochargesappliedifrequestedbyr+eusers\\"needed?'] = "no_charges_if_r_e_requested"

    def get_charging(db_string):
        if db_string[0] == '[':
            db_string = json.loads(db_string)[0]
        cleaned_str = db_string.strip('"').replace(" ", "").replace("-", "").replace("/", "").lower()
        key = simple_charging[cleaned_str]
        return CommercialCharges[key]

    collab = recursive_query(16652)
    collab = {(nren_name, year): get_charging(answer) for answer_id, nren_name, year, answer in collab}
    services = recursive_query(16653)
    services = {(nren_name, year): get_charging(answer) for answer_id, nren_name, year, answer in services}
    peering = recursive_query(16654)
    peering = {(nren_name, year): get_charging(answer) for answer_id, nren_name, year, answer in peering}

    for nren_name, year in collab.keys() | services.keys() | peering.keys():
        if nren_name not in nren_dict:
            continue

        nren = nren_dict[nren_name]

        commercial_charging_levels = {}

        collaboration = collab.get((nren_name, year))
        peer = peering.get((nren_name, year))
        service = services.get((nren_name, year))

        if collaboration:
            commercial_charging_levels['collaboration'] = {
                'charging_level': collaboration.name
            }
        if peer:
            commercial_charging_levels['peering'] = {
                'charging_level': peer.name
            }
        if service:
            commercial_charging_levels['services'] = {
                'charging_level': service.name
            }

        yield ('commercial_charging_levels', nren, nren.id, year, commercial_charging_levels)


def fibre_light(nren_dict):
    comment_map = mapping.VALUE_TO_CODE_MAPPING.get(16668)

    fibre = recursive_query(16668)
    fibre = {(nren_name, year): answer.strip('"') for answer_id, nren_name, year, answer in fibre}
    fibre_comment = recursive_query(16669)
    fibre_comment = {(nren_name, year): answer.strip('"') for answer_id, nren_name, year, answer in fibre_comment}

    for nren_name, year in fibre.keys() | fibre_comment.keys():
        if nren_name not in nren_dict:
            continue

        description = fibre.get((nren_name, year))
        comment = fibre_comment.get((nren_name, year))
        if description and description[0:5] != "Other":
            if comment and comment.replace("-", "") != "":
                pass  # previously used to log comments, just skip them for now
        else:
            description = comment

        if description:
            is_other = description not in comment_map
            description = comment_map.get(description, description).replace("\\", "")
            if is_other:
                yield ('fibre_light', nren_dict[nren_name], nren_dict[nren_name].id, year, "other")
                yield ('fibre_light-Comment', nren_dict[nren_name], nren_dict[nren_name].id, year, description)
            else:
                yield ('fibre_light', nren_dict[nren_name], nren_dict[nren_name].id, year, description)


def network_map_urls(nren_dict):

    rows = recursive_query(16670)

    for answer_id, nren_name, year, answer in rows:
        if nren_name not in nren_dict:
            continue

        urls = extract_urls(text=answer)
        if not urls:
            continue

        network_map_urls = []

        for url in urls:
            network_map_urls.append({'network_map_url': url})

        yield ('network_map_urls', nren_dict[nren_name], nren_dict[nren_name].id, year, network_map_urls)


def monitoring_tools(nren_dict):
    description_map = mapping.VALUE_TO_CODE_MAPPING.get(16672)

    tools = recursive_query(16672)
    tools = {(nren_name, year): json.loads(answer) for answer_id, nren_name, year, answer in tools}
    tools_comment = recursive_query(16673)
    tools_comment = {(nren_name, year): answer.strip('" ') for answer_id, nren_name, year, answer in tools_comment}
    netflow = recursive_query(16674)
    netflow = {(nren_name, year): answer.strip('" ') for answer_id, nren_name, year, answer in netflow}

    for nren_name, year in tools.keys() | tools_comment.keys() | netflow.keys():
        if nren_name not in nren_dict:
            continue

        nren = nren_dict[nren_name]

        tool_descriptions = tools.get((nren_name, year), [])
        comment = tools_comment.get((nren_name, year), "").replace("-", "")
        if comment:
            tool_descriptions.append(comment)
            if "Other" in tool_descriptions:
                tool_descriptions.remove("Other")
            if "Other " in tool_descriptions:
                tool_descriptions.remove("Other ")

        monitoring_tools = []
        monitoring_tools_comment = ""
        other_tools = []

        for description in tool_descriptions:
            description = description.strip()
            if not description:
                continue
            if description in description_map:
                monitoring_tools.append(description_map[description])
            else:
                other_tools.append(description)

        if other_tools:
            monitoring_tools.append("other")
            monitoring_tools_comment = ", ".join(other_tools)

        yield ('monitoring_tools', nren, nren.id, year, monitoring_tools)

        if monitoring_tools_comment:
            yield ('monitoring_tools-Comment', nren, nren.id, year, monitoring_tools_comment)

        # netflow processing description
        if netflow.get((nren_name, year)):
            yield ('netflow_vendors', nren, nren.id, year, netflow.get((nren_name, year)))


def traffic_statistics(nren_dict):

    stats = recursive_query(16677)
    stat_urls = recursive_query(16678)
    stat_urls = {(nren_name, year): answer for answer_id, nren_name, year, answer in stat_urls}

    for answer_id, nren_name, year, answer in stats:
        if nren_name not in nren_dict:
            continue

        nren = nren_dict[nren_name]

        db_urls = stat_urls.get((nren_name, year))
        if db_urls:
            urls = extract_urls(text=db_urls)
            db_urls = urls
        else:
            db_urls = []

        valid_urls = []
        for url in db_urls:
            if valid_url(url):
                valid_urls.append(url)

        if not valid_urls:
            continue

        yield ('traffic_statistics', nren, nren.id, year, "Yes")
        yield ('traffic_statistics_urls', nren, nren.id, year, [{"traffic_statistics_url": url} for url in valid_urls])


def siem_vendors(nren_dict):

    vendors = recursive_query(16679)
    vendors = {(nren_name, year): json.loads(answer) for answer_id, nren_name, year, answer in vendors}
    vendor_comment = recursive_query(16680)
    vendor_comment = {(nren_name, year): answer.strip('"') for answer_id, nren_name, year, answer in vendor_comment}

    for nren_name, year in vendors.keys() | vendor_comment.keys():
        if nren_name not in nren_dict:
            continue

        nren = nren_dict[nren_name]

        vendor_names = vendors.get((nren_name, year), [])
        comment = vendor_comment.get((nren_name, year))
        if comment:
            vendor_names.remove("Other")

        siem_soc_vendor = [*vendor_names]
        siem_soc_vendor_comment = ""
        if comment:
            siem_soc_vendor.append("other")
            siem_soc_vendor_comment = comment

        yield ('siem_soc_vendor', nren, nren.id, year, siem_soc_vendor)
        if siem_soc_vendor_comment:
            yield ('siem_soc_vendor-Comment', nren, nren.id, year, siem_soc_vendor_comment)


def certificate_providers(nren_dict):

    provider_map = mapping.VALUE_TO_CODE_MAPPING.get(16681)

    providers = recursive_query(16681)
    providers = {(nren_name, year): json.loads(answer) for answer_id, nren_name, year, answer in providers}
    prov_comment = recursive_query(16682)
    prov_comment = {(nren_name, year): answer.strip('"') for answer_id, nren_name, year, answer in prov_comment}

    for nren_name, year in providers.keys() | prov_comment.keys():
        if nren_name not in nren_dict:
            continue

        nren = nren_dict[nren_name]

        provider_names = providers.get((nren_name, year), [])
        comment = prov_comment.get((nren_name, year))
        if comment:
            provider_names.append(comment)
            if "Other" in provider_names:
                provider_names.remove("Other")

        def _replace_provider(provider):
            if 'let' in provider.lower() and 'encrypt' in provider.lower():
                return "Let's Encrypt"
            return provider_map.get(provider, provider)

        provider_names = [_replace_provider(p) for p in provider_names]

        certificate_service = []
        certificate_service_comment = []

        for provider in provider_names:
            if provider in provider_map:
                certificate_service.append(provider)
            else:
                if "other" not in certificate_service:
                    certificate_service.append("other")
                certificate_service_comment.append(provider)

        yield ('certificate_service', nren, nren.id, year, certificate_service)
        if certificate_service_comment:
            yield ('certificate_service-Comment', nren, nren.id, year, ", ".join(certificate_service_comment))


def weather_map(nren_dict):

    weather = recursive_query(16683)
    urls = recursive_query(16684)
    urls = {(nren_name, year): answer.strip('" ') for answer_id, nren_name, year, answer in urls}

    for answer_id, nren_name, year, answer in weather:
        if nren_name not in nren_dict:
            continue

        url = urls.get((nren_name, year), "")
        if url:
            found_urls = extract_urls(text=url)
            if found_urls:
                url = found_urls[0]
            else:
                url = ""

        valid = valid_url(url)

        if not valid:
            continue

        yield ('network_weather', nren_dict[nren_name], nren_dict[nren_name].id, year, "Yes")
        yield ('network_weather_url', nren_dict[nren_name], nren_dict[nren_name].id, year, url)


def pert_team(nren_dict):

    rows = recursive_query(16685)

    for answer_id, nren_name, year, answer in rows:
        if nren_name not in nren_dict:
            continue

        if answer == "null":
            continue
        pert = YesNoPlanned[answer.strip('"').lower()]

        pert = pert.name[0].upper() + pert.name[1:]

        yield ('pert_team', nren_dict[nren_name], nren_dict[nren_name].id, year, pert)


def alien_wave(nren_dict):

    alien = recursive_query(16687)
    alien = {
        (nren_name, year): YesNoPlanned[answer.strip('"').lower()] for answer_id, nren_name, year, answer in alien
    }
    nr = recursive_query(16688)
    nr = {(nren_name, year): int(answer.strip('"')) for answer_id, nren_name, year, answer in nr}
    internal = recursive_query(16689)
    internal = {(nren_name, year): answer == '"Yes"' for answer_id, nren_name, year, answer in internal}

    for nren_name, year in alien.keys() | nr.keys() | internal.keys():
        if nren_name not in nren_dict:
            continue

        nren = nren_dict[nren_name]

        alien_wave_third_party = alien.get((nren_name, year))
        nr_of_alien_wave_third_party_services = nr.get((nren_name, year))
        alien_wave_internal = internal.get((nren_name, year))

        if alien_wave_third_party is not None or nr_of_alien_wave_third_party_services is not None:
            name = (alien_wave_third_party or YesNoPlanned.yes).name  # default to yes if there's services, but no flag
            name = name[0].upper() + name[1:]
            yield ('alienwave_services', nren, nren.id, year, name)

        if nr_of_alien_wave_third_party_services is not None:
            yield ('alienwave_services_number', nren, nren.id, year, nr_of_alien_wave_third_party_services)

        if alien_wave_internal is not None:
            yield ('alienwave_internal', nren, nren.id, year, "Yes" if alien_wave_internal else "No")


def external_connections(nren_dict):

    question_nrs = {
        16694: (5, "capacity"),
        16695: (7, "capacity"),
        16696: (6, "capacity"),
        16697: (7, "from_organization"),
        16698: (1, "to_organization"),
        16699: (8, "to_organization"),
        16700: (9, "to_organization"),
        16701: (1, "from_organization"),
        16702: (8, "capacity"),
        16703: (5, "to_organization"),
        16704: (0, "link_name"),
        16705: (1, "link_name"),
        16706: (9, "capacity"),
        16707: (2, "link_name"),
        16708: (0, "from_organization"),
        16709: (4, "link_name"),
        16710: (3, "link_name"),
        16711: (9, "link_name"),
        16712: (7, "link_name"),
        16713: (8, "link_name"),
        16714: (6, "link_name"),
        16715: (5, "link_name"),
        16716: (4, "from_organization"),
        16717: (5, "from_organization"),
        16718: (6, "from_organization"),
        16719: (2, "to_organization"),
        16720: (3, "to_organization"),
        16721: (4, "to_organization"),
        16722: (6, "to_organization"),
        16723: (7, "to_organization"),
        16724: (2, "interconnection_method"),
        16725: (3, "interconnection_method"),
        16726: (4, "interconnection_method"),
        16727: (5, "interconnection_method"),
        16728: (8, "from_organization"),
        16729: (9, "from_organization"),
        16730: (0, "to_organization"),
        16731: (0, "capacity"),
        16732: (1, "capacity"),
        16733: (2, "capacity"),
        16734: (3, "capacity"),
        16735: (4, "capacity"),
        16736: (3, "from_organization"),
        16737: (2, "from_organization"),
        16738: (1, "interconnection_method"),
        16739: (7, "interconnection_method"),
        16740: (8, "interconnection_method"),
        16741: (0, "interconnection_method"),
        16742: (9, "interconnection_method"),
        16743: (6, "interconnection_method")
    }

    connection_dicts = {}
    nren_year_set = set()
    for question_id, (connection_nr, field) in question_nrs.items():
        rows = recursive_query(question_id)
        for answer_id, nren_name, year, answer in rows:
            nren_year_set.add((nren_name, year))
            conn_dict = connection_dicts.setdefault((nren_name, year, connection_nr), {})
            conn_dict[field] = answer.strip('" ')

    int_simple = {key.replace(" ", "").lower(): value for key, value in INTERCONNECTION.items()}
    int_simple['openexchangepoi'] = "open_exchange"

    for conn_dict in connection_dicts.values():
        if conn_dict.get('capacity'):
            try:
                conn_dict['capacity'] = str(Decimal(conn_dict['capacity'].split('G')[0].strip()))
            except:  # noqa: E722
                # Capacity could not be converted to a number
                conn_dict['capacity'] = None
        if conn_dict.get('interconnection_method'):
            int_conn = int_simple[conn_dict['interconnection_method'].replace(" ", "").lower()]
            conn_dict['interconnection_method'] = ConnectionMethod[int_conn].name

    for nren_name, year in nren_year_set:
        if nren_name not in nren_dict:
            continue

        connections = []
        for connection_nr in range(0, 10):
            conn = connection_dicts.get((nren_name, year, connection_nr))
            if conn:
                connections.append(conn)

        yield ('external_connections', nren_dict[nren_name], nren_dict[nren_name].id, year, connections)


def network_automation(nren_dict):

    network_automation_map = mapping.VALUE_TO_CODE_MAPPING.get(16758)

    rows = recursive_query(16757)
    tasks = recursive_query(16758)
    tasks = {(nren_name, year): json.loads(answer) for answer_id, nren_name, year, answer in tasks}

    for answer_id, nren_name, year, answer in rows:
        if nren_name not in nren_dict:
            continue

        nren = nren_dict[nren_name]

        network_automation = YesNoPlanned[answer.strip('"').lower()]
        specifics = tasks.get((nren_name, year), [])

        network_automation_tasks = []

        for task in specifics:
            if task in network_automation_map:
                # we don't allow "other" in the tasks in the survey, so only map the known tasks
                network_automation_tasks.append(network_automation_map[task])
        name = network_automation.name
        name = name[0].upper() + name[1:]
        yield ('network_automation', nren, nren.id, year, name)

        yield ('network_automation_tasks', nren, nren.id, year, network_automation_tasks)


def network_function_virtualisation(nren_dict):

    nfv_map = mapping.VALUE_TO_CODE_MAPPING.get(16755)
    rows = recursive_query(16754)
    types = recursive_query(16755)
    types = {(nren_name, year): json.loads(answer) for answer_id, nren_name, year, answer in types}
    types_comment = recursive_query(16756)
    types_comment = {(nren_name, year): answer.strip('" ') for answer_id, nren_name, year, answer in types_comment}

    for answer_id, nren_name, year, answer in rows:
        if nren_name not in nren_dict:
            continue

        nren = nren_dict[nren_name]

        nfv = YesNoPlanned[answer.strip('"').lower()]
        nfv = nfv.name[0].upper() + nfv.name[1:]
        specifics = types.get((nren_name, year), [])
        specifics = list(itertools.chain(*[s.split(', ') for s in specifics if s]))
        comment = types_comment.get((nren_name, year), "").replace("-", "")
        if comment:
            specifics.append(comment)
            if "Other" in specifics:
                specifics.remove("Other")

        nfv_types = []
        nfv_types_comment = []
        for task in specifics:
            if task in nfv_map:
                nfv_types.append(nfv_map.get(task))
            else:
                nfv_types_comment.append(task)
                if "other" not in nfv_types:
                    nfv_types.append("other")

        yield ('nfv', nren, nren.id, year, nfv)

        if nfv_types:
            yield ('nfv_types', nren, nren.id, year, nfv_types)

        if nfv_types_comment:
            yield ('nfv_types-Comment', nren, nren.id, year, ", ".join(nfv_types_comment))


def fetch_data():
    # requires being in a flask app context when called
    nren_dict = helpers.get_uppercase_nren_dict()
    yield from budget(nren_dict)
    yield from funding_sources(nren_dict)
    yield from staff_data(nren_dict)
    yield from nren_parent_org(nren_dict)
    yield from nren_sub_org(nren_dict)
    yield from charging_structure(nren_dict)
    yield from ec_projects(nren_dict)
    yield from policies(nren_dict)
    yield from institutions_urls(nren_dict)

    yield from central_procurement(nren_dict)
    yield from service_management(nren_dict)
    yield from service_user_types(nren_dict)
    yield from standards(nren_dict)
    yield from crisis_exercises(nren_dict)
    yield from security_controls(nren_dict)
    yield from commercial_connectivity(nren_dict)
    yield from commercial_charging_level(nren_dict)

    yield from fibre_light(nren_dict)
    yield from network_map_urls(nren_dict)
    yield from monitoring_tools(nren_dict)
    yield from traffic_statistics(nren_dict)
    yield from siem_vendors(nren_dict)
    yield from certificate_providers(nren_dict)
    yield from weather_map(nren_dict)
    yield from pert_team(nren_dict)
    yield from alien_wave(nren_dict)
    yield from external_connections(nren_dict)
    yield from network_function_virtualisation(nren_dict)
    yield from network_automation(nren_dict)
