import logging
from decimal import Decimal
from typing import List, Dict, Any, Type
from collections import defaultdict
from compendium_v2.db.presentation_models import BudgetEntry, ChargingStructure, ECProject, ExternalConnections, \
    InstitutionURLs, NrenStaff, ParentOrganization, Policy, SubOrganization, TrafficVolume, \
    FundingSource, CentralProcurement, ServiceManagement, ServiceUserTypes, EOSCListings, \
    Standards, CrisisExercises, SecurityControls, ConnectedProportion, ConnectivityLevel, \
    ConnectionCarrier, ConnectivityLoad, ConnectivityGrowth, CommercialConnectivity, \
    CommercialChargingLevel, RemoteCampuses, DarkFibreLease, DarkFibreInstalled, FibreLight, \
    NetworkMapUrls, MonitoringTools, PassiveMonitoring, TrafficStatistics, SiemVendors, \
    CertificateProviders, WeatherMap, PertTeam, AlienWave, Capacity, NonREPeers, TrafficRatio, \
    OpsAutomation, NetworkFunctionVirtualisation, NetworkAutomation, NRENService, NREN, PresentationModel
from compendium_v2.db.presentation_model_enums import CarryMechanism, CommercialCharges, YesNoPlanned, \
    CommercialConnectivityCoverage, ConnectivityCoverage, FeeType, MonitoringMethod, ServiceCategory, UserCategory
from compendium_v2.db.survey_models import SurveyResponse
from compendium_v2.publishers.utils import float_or_zero, float_or_none, bool_or_none, int_or_none, sanitize
from compendium_v2.publishers.helpers import valid_url, merge_string

logger = logging.getLogger(__name__)


def map_budget_entry(year: int, nren: NREN, answers: Dict[str, Any]):
    budget = answers.get("budget")
    if budget:
        # replace comma with dot for decimal, surveyjs allows commas for decimals (common in EU countries)
        budget = budget.replace(",", ".")
        return {
            'nren_id': nren.id,
            'year': year,
            'budget': Decimal(budget)
        }


def map_fundingsource_entry(year: int, nren: NREN, answers: Dict[str, Any]):
    funding_source = answers.get("income_sources")
    if funding_source:
        client_institutions = float_or_zero(funding_source, "client_institutions")
        european_funding = float_or_zero(funding_source, "european_funding")
        gov_public_bodies = float_or_zero(funding_source, "gov_public_bodies")
        commercial = float_or_zero(funding_source, "commercial")
        other = float_or_zero(funding_source, "other")

        if sum([client_institutions, european_funding, gov_public_bodies, commercial, other]) == 0:
            return None

        return {
            'nren_id': nren.id,
            'year': year,
            'client_institutions': client_institutions,
            'european_funding': european_funding,
            'gov_public_bodies': gov_public_bodies,
            'commercial': commercial,
            'other': other
        }


def map_charging_structure_entry(year: int, nren: NREN, answers: Dict[str, Any]):
    charging = answers.get("charging_mechanism")
    if charging:
        return {
            'nren_id': nren.id,
            'year': year,
            'fee_type': FeeType[charging]
        }


def map_nren_staff(year: int, nren: NREN, answers: Dict[str, Any]):
    staff_roles = answers.get("staff_roles", {})
    staff_employment_type = answers.get("staff_employment_type", {})
    if staff_roles or staff_employment_type:
        return {
            'nren_id': nren.id,
            'year': year,
            'permanent_fte': float_or_zero(staff_employment_type, "permanent_fte"),
            'subcontracted_fte': float_or_zero(staff_employment_type, "subcontracted_fte"),
            'technical_fte': float_or_zero(staff_roles, "technical_fte"),
            'non_technical_fte': float_or_zero(staff_roles, "nontechnical_fte")
        }


def map_parent_orgs(nren: NREN, year: int, answers: Dict[str, Any]):
    has_parent = answers.get("parent_organization") == "Yes"
    parent = answers.get("parent_organization_name", "").strip()
    if len(parent) > 256:
        logger.warning(f"Parent organization name too long for NREN {nren.id}")
        parent = parent[:256]
    if has_parent and parent:
        return {
            'nren_id': nren.id,
            'year': year,
            'organization': parent
        }


def map_sub_orgs(nren: NREN, year: int, answers: Dict[str, Any]):
    has_subs = answers.get("suborganizations") == 'Yes'
    subs = answers.get("suborganization_details")
    if has_subs and subs:
        for sub in subs:
            role = sub.get("suborganization_role", "").strip()
            if role == "other":
                role = sub.get("suborganization_role-Comment", "other").strip()
            name = sub.get("suborganization_name", "").strip()
            if not name:
                continue
            if len(name) > 256:
                logger.warning(f"Suborganization name too long for NREN {nren.name} ({year})")
                logger.warning(name)
                name = name[:256]
            if len(role) > 256:
                logger.warning(f"Organization role too long for NREN {nren.name} ({year})")
                logger.warning(role)
                role = role[:256]
            yield {
                'nren_id': nren.id,
                'year': year,
                'organization': name,
                'role': role
            }


def map_ec_projects(nren: NREN, year: int, answers: Dict[str, Any]):
    has_ec_projects = answers.get("ec_projects") == "Yes"
    ec_projects = answers.get("ec_project_names")
    if has_ec_projects and ec_projects:
        for ec_project in ec_projects:
            if ec_project:
                yield {
                    'nren_id': nren.id,
                    'year': year,
                    'project': ec_project.get("ec_project_name")
                }


def map_policy(nren: NREN, year: int, answers: Dict[str, Any]):
    strategy = answers.get("corporate_strategy_url", "").strip()
    if strategy and not valid_url(strategy):
        strategy = ""

    policies = answers.get("policies", {})

    environmental_url = policies.get("environmental_policy", {}).get("url", "").strip()
    if environmental_url and not valid_url(environmental_url):
        environmental_url = ""

    equal_opportunity_url = policies.get("equal_opportunity_policy", {}).get("url", "").strip()
    if equal_opportunity_url and not valid_url(equal_opportunity_url):
        equal_opportunity_url = ""

    connectivity_url = policies.get("connectivity_policy", {}).get("url", "").strip()
    if connectivity_url and not valid_url(connectivity_url):
        connectivity_url = ""

    acceptable_use_url = policies.get("acceptable_use_policy", {}).get("url", "").strip()
    if acceptable_use_url and not valid_url(acceptable_use_url):
        acceptable_use_url = ""

    privacy_notice_url = policies.get("privacy_notice", {}).get("url", "").strip()
    if privacy_notice_url and not valid_url(privacy_notice_url):
        privacy_notice_url = ""

    data_protection_url = policies.get("data_protection_contact", {}).get("url", "").strip()
    if data_protection_url and not valid_url(data_protection_url):
        data_protection_url = ""

    gender_equality_url = policies.get("gender_equality_policy", {}).get("url", "").strip()
    if gender_equality_url and not valid_url(gender_equality_url):
        gender_equality_url = ""

    all_policies = [strategy, environmental_url, equal_opportunity_url, connectivity_url,
                    acceptable_use_url, privacy_notice_url, data_protection_url, gender_equality_url]
    if any(all_policies):
        return {
            'nren_id': nren.id,
            'year': year,
            'strategic_plan': strategy,
            'environmental': environmental_url,
            'equal_opportunity': equal_opportunity_url,
            'connectivity': connectivity_url,
            'acceptable_use': acceptable_use_url,
            'privacy_notice': privacy_notice_url,
            'data_protection': data_protection_url,
            'gender_equality': gender_equality_url
        }


def map_traffic_volume(nren: NREN, year: int, answers: Dict[str, Any]):
    traffic_estimate = answers.get("traffic_estimate")
    if traffic_estimate:
        return {
            'nren_id': nren.id,
            'year': year,
            'to_customers': float_or_zero(traffic_estimate, "to_customers"),
            'from_customers': float_or_zero(traffic_estimate, "from_customers"),
            'to_external': float_or_zero(traffic_estimate, "to_external"),
            'from_external': float_or_zero(traffic_estimate, "from_external")
        }


def map_institution_urls(nren: NREN, year: int, answers: Dict[str, Any]):
    institution_urls = answers.get("connected_sites_lists")
    if institution_urls:
        urls = []

        for u in institution_urls:
            url = u.get("connected_sites_url", "").strip()
            if url and valid_url(url):
                urls.append(url)

        if urls:
            return {
                'nren_id': nren.id,
                'year': year,
                'urls': urls
            }


def map_central_procurement(nren: NREN, year: int, answers: Dict[str, Any]):
    central_procurement = answers.get("central_software_procurement") == "Yes"
    if central_procurement:
        central_procurement_amount = float_or_none(answers, "central_procurement_amount")
    else:
        central_procurement_amount = None
    return {
        'nren_id': nren.id,
        'year': year,
        'central_procurement': central_procurement,
        'amount': central_procurement_amount
    }


def map_service_management(nren: NREN, year: int, answers: Dict[str, Any]):
    formal_service_management_framework = answers.get("formal_service_management_framework")
    service_level_targets = answers.get("service_level_targets")
    if formal_service_management_framework or service_level_targets:
        return {
            'nren_id': nren.id,
            'year': year,
            'service_management_framework': bool_or_none(formal_service_management_framework),
            'service_level_targets': bool_or_none(service_level_targets)
        }


def map_service_user_types(nren: NREN, year: int, answers: Dict[str, Any]):
    service_type_matrix = answers.get("service_matrix", {})
    for user_type, service_types in service_type_matrix.items():
        user_type_enum = UserCategory[user_type]
        service_types_list = service_types["service_types"]
        for service_type in service_types_list:
            service_type_enum = ServiceCategory[service_type]
            yield {
                'nren_id': nren.id,
                'year': year,
                'user_category': user_type_enum,
                'service_category': service_type_enum
            }


def map_eosc_listings(nren: NREN, year: int, answers: Dict[str, Any]):
    has_eosc_listings = answers.get("service_portfolio_eosc_portal") == "Yes"
    services_on_eosc_portal_list = answers.get("services_on_eosc_portal_list", [])
    eosc_list = [i.get("service_name") for i in services_on_eosc_portal_list if i.get("service_name")]
    if has_eosc_listings and eosc_list:
        return {
            'nren_id': nren.id,
            'year': year,
            'service_names': eosc_list
        }


def map_standards(nren: NREN, year: int, answers: Dict[str, Any]):
    audits = bool_or_none(answers.get("audits"))
    audits_specifics = answers.get("audit_specifics", "").strip()
    business_continuity_plans = bool_or_none(answers.get("business_continuity_plans"))
    bcp_specifics = answers.get("business_continuity_plans_specifics", "").strip()
    crisis_management_procedure = bool_or_none(answers.get("crisis_management_procedure"))

    if all([p is None or p == '' for p in [audits, audits_specifics, business_continuity_plans,
                                           bcp_specifics, crisis_management_procedure]]):
        return None
    return {
        'nren_id': nren.id,
        'year': year,
        'audits': audits,
        'audit_specifics': audits_specifics,
        'business_continuity_plans': business_continuity_plans,
        'business_continuity_plans_specifics': bcp_specifics,
        'crisis_management_procedure': crisis_management_procedure
    }


def map_crisis_exercises(nren: NREN, year: int, answers: Dict[str, Any]):
    crisis_exercises = answers.get("crisis_exercises")
    if crisis_exercises:
        return {
            'nren_id': nren.id,
            'year': year,
            'exercise_descriptions': crisis_exercises
        }


def map_security_controls(nren: NREN, year: int, answers: Dict[str, Any]):
    security_controls = answers.get("security_controls")
    if security_controls:
        if "other" in security_controls:
            security_controls.remove("other")
            comment = answers.get("security_controls-Comment", "").strip()
            if comment:
                security_controls.append(comment)
        return {
            'nren_id': nren.id,
            'year': year,
            'security_control_descriptions': security_controls
        }


def map_connected_proportion(nren: NREN, year: int, answers: Dict[str, Any]):
    connectivity_proportions = answers.get("connectivity_proportions", {})
    for user_type, connectivity_proportion in connectivity_proportions.items():
        user_type_enum = UserCategory[user_type]
        coverage = connectivity_proportion.get("covered")
        coverage_enum = ConnectivityCoverage[coverage] if coverage else None
        number_connected = int_or_none(connectivity_proportion, "nr_connected")
        market_share = float_or_none(connectivity_proportion, "market_share_percentage")
        users_served = int_or_none(connectivity_proportion, "nr_of_users")
        yield {
            'nren_id': nren.id,
            'year': year,
            'user_category': user_type_enum,
            'coverage': coverage_enum,
            'number_connected': number_connected,
            'market_share': market_share,
            'users_served': users_served
        }


def map_connectivity_levels(nren: NREN, year: int, answers: Dict[str, Any]):
    connectivity_levels = answers.get("connectivity_level", {})
    for user_type, connectivity_level in connectivity_levels.items():
        user_type_enum = UserCategory[user_type]
        typical_speed = int_or_none(connectivity_level, "typical_speed")
        highest_speed = int_or_none(connectivity_level, "highest_speed")
        highest_speed_proportion = float_or_none(connectivity_level, "highest_speed_connection_percentage")
        yield {
            'nren_id': nren.id,
            'year': year,
            'user_category': user_type_enum,
            'typical_speed': typical_speed,
            'highest_speed': highest_speed,
            'highest_speed_proportion': highest_speed_proportion
        }


def map_connection_carriers(nren: NREN, year: int, answers: Dict[str, Any]):
    traffic_carriers = answers.get("traffic_carriers", {})
    for user_type, traffic_carrier in traffic_carriers.items():
        user_type_enum = UserCategory[user_type]
        carry_mechanism = traffic_carrier.get("carry_mechanism", "").strip()
        if carry_mechanism:
            yield {
                'nren_id': nren.id,
                'year': year,
                'user_category': user_type_enum,
                'carry_mechanism': CarryMechanism[carry_mechanism]
            }


def map_connectivity_loads(nren: NREN, year: int, answers: Dict[str, Any]):
    traffic_loads = answers.get("traffic_load", {})
    for user_type, traffic_load in traffic_loads.items():
        user_type_enum = UserCategory[user_type]
        yield {
            'nren_id': nren.id,
            'year': year,
            'user_category': user_type_enum,
            'average_load_from_institutions': int_or_none(traffic_load, "average_from_institutions_to_network"),
            'average_load_to_institutions': int_or_none(traffic_load, "average_to_institutions_from_network"),
            'peak_load_from_institutions': int_or_none(traffic_load, "peak_from_institutions_to_network"),
            'peak_load_to_institutions': int_or_none(traffic_load, "peak_to_institutions_from_network")
        }


def map_connectivity_growth(nren: NREN, year: int, answers: Dict[str, Any]):
    traffic_growths = answers.get("traffic_growth", {})
    for user_type, traffic_growth in traffic_growths.items():
        user_type_enum = UserCategory[user_type]
        yield {
            'nren_id': nren.id,
            'year': year,
            'user_category': user_type_enum,
            'growth': float_or_zero(traffic_growth, "growth_rate")
        }


def map_commercial_connectivity(nren: NREN, year: int, answers: Dict[str, Any]):
    commercial_organizations = answers.get("commercial_organizations")
    if commercial_organizations:
        c1 = commercial_organizations.get("commercial_r_e", {}).get("connection")
        c2 = commercial_organizations.get("commercial_general", {}).get("connection")
        c3 = commercial_organizations.get("commercial_collaboration", {}).get("connection")
        c4 = commercial_organizations.get("commercial_service_provider", {}).get("connection")
        c5 = commercial_organizations.get("university_spin_off", {}).get("connection")
        yield {
            'nren_id': nren.id,
            'year': year,
            'commercial_r_and_e': CommercialConnectivityCoverage[c1] if c1 else None,
            'commercial_general': CommercialConnectivityCoverage[c2] if c2 else None,
            'commercial_collaboration': CommercialConnectivityCoverage[c3] if c3 else None,
            'commercial_service_provider': CommercialConnectivityCoverage[c4] if c4 else None,
            'university_spin_off': CommercialConnectivityCoverage[c5] if c5 else None,
        }


def map_commercial_charging_levels(nren: NREN, year: int, answers: Dict[str, Any]):
    commercial_charging_levels = answers.get("commercial_charging_levels")
    if commercial_charging_levels:
        c1 = commercial_charging_levels.get("collaboration", {}).get("charging_level")
        c2 = commercial_charging_levels.get("services", {}).get("charging_level")
        c3 = commercial_charging_levels.get("peering", {}).get("charging_level")
        yield {
            'nren_id': nren.id,
            'year': year,
            'collaboration': CommercialCharges[c1] if c1 else None,
            'service_supplier': CommercialCharges[c2] if c2 else None,
            'direct_peering': CommercialCharges[c3] if c3 else None,
        }


def map_remote_campuses(nren: NREN, year: int, answers: Dict[str, Any]):
    has_remote_campuses = answers.get("remote_campuses")
    if has_remote_campuses:
        remote_campuses = has_remote_campuses == "Yes"
        remote_campuses_specifics = []
        specifics = answers.get("remote_campuses_specifics", [])
        if remote_campuses or specifics:
            remote_campuses_specifics = [
                {"country": i.get("country", "").strip(), "local_r_and_e_connection": bool_or_none(i, "connected")}
                for i in specifics if len(i.get("country", "").strip()) > 1  # ignore empty country
            ]
        return {
            'nren_id': nren.id,
            'year': year,
            'remote_campus_connectivity': remote_campuses,
            'connections': remote_campuses_specifics
        }


def map_dark_fibre_lease(nren: NREN, year: int, answers: Dict[str, Any]):
    iru_or_lease = answers.get("dark_fibre_lease")
    iru_duration = float_or_none(answers, "dark_fibre_lease_duration")
    inside_country = int_or_none(answers, "dark_fibre_lease_kilometers_inside_country")
    outside_country = int_or_none(answers, "dark_fibre_lease_kilometers_outside_country")

    if not any([p is not None for p in [iru_or_lease, iru_duration, inside_country, outside_country]]):
        return None
    return {
        'nren_id': nren.id,
        'year': year,
        'iru_or_lease': iru_or_lease == "Yes",
        'fibre_length_in_country': inside_country,
        'fibre_length_outside_country': outside_country,
        'iru_duration': iru_duration
    }


def map_dark_fibre_installed(nren: NREN, year: int, answers: Dict[str, Any]):
    installed = answers.get("dark_fibre_nren")
    length = int_or_none(answers, "dark_fibre_nren_kilometers_inside_country")
    if not any([p is not None for p in [installed, length]]):
        return None
    return {
        'nren_id': nren.id,
        'year': year,
        'installed': installed == "Yes",
        'fibre_length_in_country': length if installed == "Yes" else None
    }


def map_fibre_light(nren: NREN, year: int, answers: Dict[str, Any]):
    fibre_light = answers.get("fibre_light")
    if fibre_light:
        if fibre_light == "other":
            fibre_light = answers.get("fibre_light-Comment", "other").strip()
        return {
            'nren_id': nren.id,
            'year': year,
            'light_description': fibre_light
        }


def map_monitoring_tools(nren: NREN, year: int, answers: Dict[str, Any]):
    monitoring_tools = answers.get("monitoring_tools", [])
    netflow_vendors = answers.get("netflow_vendors", "").strip()
    tools_comment = answers.get("monitoring_tools-Comment", "").strip()
    if len(tools_comment) > 1:
        if "other" in monitoring_tools:
            monitoring_tools.remove("other")
        monitoring_tools.append(tools_comment)
    if monitoring_tools or netflow_vendors:
        return {
            'nren_id': nren.id,
            'year': year,
            'tool_descriptions': monitoring_tools,
            'netflow_processing_description': netflow_vendors
        }


def map_passive_monitoring(nren: NREN, year: int, answers: Dict[str, Any]):
    passive_monitoring = answers.get("passive_monitoring")
    if passive_monitoring:
        passive_monitoring = passive_monitoring == "Yes"
        passive_monitoring_tech = answers.get("passive_monitoring_tech")
        return {
            'nren_id': nren.id,
            'year': year,
            'monitoring': passive_monitoring,
            'method': MonitoringMethod[passive_monitoring_tech] if passive_monitoring_tech else None
        }


def map_traffic_statistics(nren: NREN, year: int, answers: Dict[str, Any]):
    traffic_statistics = answers.get("traffic_statistics")
    if traffic_statistics:
        traffic_statistics = traffic_statistics == "Yes"
        urls = []
        for answer in answers.get("traffic_statistics_urls", []):
            url = answer.get("traffic_statistics_url", "").strip()
            if url and valid_url(url):
                urls.append(url)

        return {
            'nren_id': nren.id,
            'year': year,
            'traffic_statistics': traffic_statistics,
            'urls': urls
        }


def map_siem_vendors(nren: NREN, year: int, answers: Dict[str, Any]):
    siem_soc_vendor = answers.get("siem_soc_vendor")
    if siem_soc_vendor:
        if "other" in siem_soc_vendor:
            siem_soc_vendor.remove("other")
            siem_soc_vendor.append(answers.get("siem_soc_vendor-Comment", "other").strip())
        return {
            'nren_id': nren.id,
            'year': year,
            'vendor_names': siem_soc_vendor
        }


def map_network_map_urls(nren: NREN, year: int, answers: Dict[str, Any]):
    network_map_urls = answers.get("network_map_urls", [])
    urls = []
    for u in network_map_urls:
        url = u.get("network_map_url", "").strip()
        if url and valid_url(url):
            urls.append(url)
    if urls:
        return {
            'nren_id': nren.id,
            'year': year,
            'urls': urls
        }


def map_certificate_providers(nren: NREN, year: int, answers: Dict[str, Any]):
    certificate_service = answers.get("certificate_service")
    if certificate_service:
        if "other" in certificate_service:
            certificate_service.remove("other")
            other_service = answers.get("certificate_service-Comment", "other").strip()

            if 'let' in other_service.lower() and 'encrypt' in other_service.lower():
                certificate_service.append("Let's Encrypt")
            else:
                certificate_service.append(other_service)
        return {
            'nren_id': nren.id,
            'year': year,
            'provider_names': certificate_service
        }


def map_weathermap_entry(nren: NREN, year: int, answers: Dict[str, Any]):
    network_weather = answers.get("network_weather")
    network_weather = network_weather == "Yes"
    if network_weather:
        url = answers.get("network_weather_url", "").strip()
        valid = valid_url(url)
        if not valid:
            return
        return {
            'nren_id': nren.id,
            'year': year,
            'weather_map': network_weather,
            'url': url
        }


def map_pert_team(nren: NREN, year: int, answers: Dict[str, Any]):
    pert_team = answers.get("pert_team")
    if pert_team:
        pert_team = YesNoPlanned[pert_team.lower()]
        return {
            'nren_id': nren.id,
            'year': year,
            'pert_team': pert_team
        }


def map_alienwave_entry(nren: NREN, year: int, answers: Dict[str, Any]):
    alienwave_services = answers.get("alienwave_services")
    alienwave_internal = answers.get("alienwave_internal")
    if alienwave_services is not None or alienwave_internal is not None:
        alienwave_services = YesNoPlanned[alienwave_services.lower()] if alienwave_services else None
        alienwave_internal = alienwave_internal == "Yes" if alienwave_internal else False
        return {
            'nren_id': nren.id,
            'year': year,
            'alien_wave_third_party': alienwave_services,
            'nr_of_alien_wave_third_party_services': int_or_none(answers, "alienwave_services_number"),
            'alien_wave_internal': alienwave_internal
        }


def map_capacity_entry(nren: NREN, year: int, answers: Dict[str, Any]):
    max_capacity = answers.get("max_capacity")
    typical_capacity = answers.get("typical_capacity")
    if max_capacity or typical_capacity:
        return {
            'nren_id': nren.id,
            'year': year,
            'largest_link_capacity': float_or_none(answers, "max_capacity"),
            'typical_backbone_capacity': float_or_none(answers, "typical_capacity")
        }


def map_external_connections(nren: NREN, year: int, answers: Dict[str, Any]):
    external_connections = answers.get("external_connections")
    if external_connections:
        connections = [{
            'link_name': connection.get('link_name', ''),
            'capacity': float(connection.get('capacity')) if connection.get('capacity') is not None else None,
            'from_organization': connection.get('from_organization', ''),
            'to_organization': connection.get('to_organization', ''),
            'interconnection_method': connection['interconnection_method']
            if connection.get('interconnection_method') is not None else None
        } for connection in external_connections]
        return {
            'nren_id': nren.id,
            'year': year,
            'connections': connections
        }


def map_non_re_peers(nren: NREN, year: int, answers: Dict[str, Any]):
    non_r_and_e_peers = answers.get("non_r_and_e_peers")
    if non_r_and_e_peers:
        return {
            'nren_id': nren.id,
            'year': year,
            'nr_of_non_r_and_e_peers': int(non_r_and_e_peers)
        }


def map_traffic_ratio(nren: NREN, year: int, answers: Dict[str, Any]):
    commodity_vs_r_e = answers.get("commodity_vs_r_e")
    if commodity_vs_r_e:
        return {
            'nren_id': nren.id,
            'year': year,
            'r_and_e_percentage': float_or_zero(commodity_vs_r_e, "r_e"),
            'commodity_percentage': float_or_zero(commodity_vs_r_e, "commodity"),
        }


def map_ops_automation(nren: NREN, year: int, answers: Dict[str, Any]):
    operational_process_automation = answers.get("operational_process_automation")
    if operational_process_automation:
        return {
            'nren_id': nren.id,
            'year': year,
            'ops_automation': YesNoPlanned[operational_process_automation.lower()],
            'ops_automation_specifics': answers.get("operational_process_automation_tools", "").strip()
        }


def map_network_function_virtualisation(nren: NREN, year: int, answers: Dict[str, Any]):
    nfv = answers.get("nfv")
    if nfv:
        nfv_types = answers.get("nfv_types", [])
        if "other" in nfv_types:
            nfv_types.remove("other")
            nfv_types.append(answers.get("nfv_types-Comment", "other").strip())
        return {
            'nren_id': nren.id,
            'year': year,
            'nfv': YesNoPlanned[nfv.lower()],
            'nfv_specifics': nfv_types
        }


def map_network_automation(nren: NREN, year: int, answers: Dict[str, Any]):
    network_automation = answers.get("network_automation")
    if network_automation:
        network_automation_tasks = answers.get("network_automation_tasks", [])
        return {
            'nren_id': nren.id,
            'year': year,
            'network_automation': YesNoPlanned[network_automation.lower()],
            'network_automation_specifics': network_automation_tasks
        }


def map_nren_services(nren: NREN, year: int, answers: Dict[str, Any]):
    all_services = {}
    for question_name in [
        "services_collaboration", "services_hosting", "services_identity", "services_isp",
        "services_multimedia", "services_network", "services_professional", "services_security"
    ]:
        all_services.update(answers.get(question_name, {}))

    optical_connection = {
        'nren_id': nren.id,
        'year': year,
        'service_key': 'dedicated-optical-connections',
    }

    for service_key, service_answers in all_services.items():

        if service_key == 'software-licenses':
            # skip software-licenses, it's basically the same responses and is merged into procurement
            continue

        # special handling for lambda and open-lightpath-exchange merging into dedicated-optical-connections
        if service_key in ('lambda', 'open-lightpath-exchange'):
            offered = service_answers.get("offered")
            if offered == ["yes"]:
                product_name = merge_string(optical_connection.get('product_name', ''),
                                            service_answers.get("name", "").strip())
                additional_information = merge_string(optical_connection.get(
                    'additional_information', ''), service_answers.get("additional_information", "").strip())
                official_description = merge_string(optical_connection.get(
                    'official_description', ''), service_answers.get("description", "").strip())
                optical_connection.update({
                    'product_name': product_name,
                    'additional_information': additional_information,
                    'official_description': official_description
                })
                continue

        offered = service_answers.get("offered")
        if offered == ["yes"]:
            yield {
                'nren_id': nren.id,
                'year': year,
                'service_key': service_key,
                'product_name': service_answers.get("name", "").strip(),
                'additional_information': service_answers.get("additional_information", "").strip(),
                'official_description': service_answers.get("description", "").strip()
            }

    if 'product_name' in optical_connection:
        yield optical_connection


def map_2023(responses: List[SurveyResponse]) -> Dict[Type[PresentationModel], List[Dict]]:
    year = 2023

    result = defaultdict(list)  # {model: [data]}

    for response in responses:
        nren = response.nren
        answers = response.answers['data']

        answers = sanitize(answers)

        budget = map_budget_entry(year, nren, answers)
        if budget:
            result[BudgetEntry].append(budget)

        funding = map_fundingsource_entry(year, nren, answers)
        if funding:
            result[FundingSource].append(funding)

        charging = map_charging_structure_entry(year, nren, answers)
        if charging:
            result[ChargingStructure].append(charging)

        nren_staff = map_nren_staff(year, nren, answers)
        if nren_staff:
            result[NrenStaff].append(nren_staff)

        parent = map_parent_orgs(nren, year, answers)
        if parent:
            result[ParentOrganization].append(parent)

        subs = map_sub_orgs(nren, year, answers)
        for sub in subs:
            result[SubOrganization].append(sub)

        ec_projects = map_ec_projects(nren, year, answers)
        for ec_project in ec_projects:
            result[ECProject].append(ec_project)

        policy = map_policy(nren, year, answers)
        if policy:
            result[Policy].append(policy)

        traffic_volume = map_traffic_volume(nren, year, answers)
        if traffic_volume:
            result[TrafficVolume].append(traffic_volume)

        institution_urls = map_institution_urls(nren, year, answers)
        if institution_urls:
            result[InstitutionURLs].append(institution_urls)

        central_procurement = map_central_procurement(nren, year, answers)
        if central_procurement:
            result[CentralProcurement].append(central_procurement)

        service_management = map_service_management(nren, year, answers)
        if service_management:
            result[ServiceManagement].append(service_management)

        service_user_types = map_service_user_types(nren, year, answers)
        for service_user_type in service_user_types:
            result[ServiceUserTypes].append(service_user_type)

        eosc_listings = map_eosc_listings(nren, year, answers)
        if eosc_listings:
            result[EOSCListings].append(eosc_listings)

        standards = map_standards(nren, year, answers)
        if standards:
            result[Standards].append(standards)

        crisis_exercises = map_crisis_exercises(nren, year, answers)
        if crisis_exercises:
            result[CrisisExercises].append(crisis_exercises)

        security_controls = map_security_controls(nren, year, answers)
        if security_controls:
            result[SecurityControls].append(security_controls)

        connected_proportions = map_connected_proportion(nren, year, answers)
        for connected_proportion in connected_proportions:
            result[ConnectedProportion].append(connected_proportion)

        connectivity_levels = map_connectivity_levels(nren, year, answers)
        for connectivity_level in connectivity_levels:
            result[ConnectivityLevel].append(connectivity_level)

        connection_carriers = map_connection_carriers(nren, year, answers)
        for connection_carrier in connection_carriers:
            result[ConnectionCarrier].append(connection_carrier)

        connectivity_loads = map_connectivity_loads(nren, year, answers)
        for connectivity_load in connectivity_loads:
            result[ConnectivityLoad].append(connectivity_load)

        connectivity_growths = map_connectivity_growth(nren, year, answers)
        for connectivity_growth in connectivity_growths:
            result[ConnectivityGrowth].append(connectivity_growth)

        commercial_connectivities = map_commercial_connectivity(nren, year, answers)
        for commercial_connectivity in commercial_connectivities:
            result[CommercialConnectivity].append(commercial_connectivity)

        commercial_charging_levels = map_commercial_charging_levels(nren, year, answers)
        for commercial_charging_level in commercial_charging_levels:
            result[CommercialChargingLevel].append(commercial_charging_level)

        remote_campuses = map_remote_campuses(nren, year, answers)
        if remote_campuses:
            result[RemoteCampuses].append(remote_campuses)

        dark_fibre_lease = map_dark_fibre_lease(nren, year, answers)
        if dark_fibre_lease:
            result[DarkFibreLease].append(dark_fibre_lease)

        dark_fibre_installed = map_dark_fibre_installed(nren, year, answers)
        if dark_fibre_installed:
            result[DarkFibreInstalled].append(dark_fibre_installed)

        fibre_light = map_fibre_light(nren, year, answers)
        if fibre_light:
            result[FibreLight].append(fibre_light)

        network_map_urls = map_network_map_urls(nren, year, answers)
        if network_map_urls:
            result[NetworkMapUrls].append(network_map_urls)

        monitoring_tools = map_monitoring_tools(nren, year, answers)
        if monitoring_tools:
            result[MonitoringTools].append(monitoring_tools)

        passive_monitoring = map_passive_monitoring(nren, year, answers)
        if passive_monitoring:
            result[PassiveMonitoring].append(passive_monitoring)

        traffic_statistics = map_traffic_statistics(nren, year, answers)
        if traffic_statistics:
            result[TrafficStatistics].append(traffic_statistics)

        siem_vendors = map_siem_vendors(nren, year, answers)
        if siem_vendors:
            result[SiemVendors].append(siem_vendors)

        certificate_providers = map_certificate_providers(nren, year, answers)
        if certificate_providers:
            result[CertificateProviders].append(certificate_providers)

        weathermap = map_weathermap_entry(nren, year, answers)
        if weathermap:
            result[WeatherMap].append(weathermap)

        pert_team = map_pert_team(nren, year, answers)
        if pert_team:
            result[PertTeam].append(pert_team)

        alienwave = map_alienwave_entry(nren, year, answers)
        if alienwave:
            result[AlienWave].append(alienwave)

        capacity = map_capacity_entry(nren, year, answers)
        if capacity:
            result[Capacity].append(capacity)

        external_connections = map_external_connections(nren, year, answers)
        if external_connections:
            result[ExternalConnections].append(external_connections)

        non_re_peers = map_non_re_peers(nren, year, answers)
        if non_re_peers:
            result[NonREPeers].append(non_re_peers)

        traffic_ratio = map_traffic_ratio(nren, year, answers)
        if traffic_ratio:
            result[TrafficRatio].append(traffic_ratio)

        ops_automation = map_ops_automation(nren, year, answers)
        if ops_automation:
            result[OpsAutomation].append(ops_automation)

        nfv = map_network_function_virtualisation(nren, year, answers)
        if nfv:
            result[NetworkFunctionVirtualisation].append(nfv)

        network_automation = map_network_automation(nren, year, answers)
        if network_automation:
            result[NetworkAutomation].append(network_automation)

        nren_services = map_nren_services(nren, year, answers)
        for nren_service in nren_services:
            result[NRENService].append(nren_service)

    return result
