"""
This module handles adding responses from the 2024 survey into the presentation_models.
The process is straightforward as the data is new and can be directly inserted into the presentation_models.

For subsequent years like 2025, handling should be almost the same as 2024, only changes in questions
will need to be addressed.

Usage:
    Used in publish_survey API in compendium_v2/routes/survey.py
"""

from typing import List, Dict, Type
from collections import defaultdict

from compendium_v2.db.survey_models import SurveyResponse
from compendium_v2.publishers.year import map_2023
from compendium_v2.publishers.utils import sanitize
from compendium_v2.db.presentation_models import BudgetEntry, ChargingStructure, ECProject, ExternalConnections, \
    InstitutionURLs, NrenStaff, ParentOrganization, Policy, SubOrganization, TrafficVolume, \
    FundingSource, CentralProcurement, ServiceManagement, ServiceUserTypes, \
    Standards, CrisisExercises, SecurityControls, ConnectedProportion, ConnectivityLevel, \
    ConnectionCarrier, ConnectivityLoad, ConnectivityGrowth, CommercialConnectivity, \
    CommercialChargingLevel, RemoteCampuses, DarkFibreLease, DarkFibreInstalled, FibreLight, \
    NetworkMapUrls, MonitoringTools, PassiveMonitoring, TrafficStatistics, SiemVendors, \
    CertificateProviders, WeatherMap, PertTeam, AlienWave, Capacity, NonREPeers, TrafficRatio, \
    OpsAutomation, NetworkFunctionVirtualisation, NetworkAutomation, NRENService, PresentationModel


def map_2024(responses: List[SurveyResponse], year=2024) -> Dict[Type[PresentationModel], List[Dict]]:
    result = defaultdict(list)  # {model: [data]}

    for response in responses:
        nren = response.nren
        answers = response.answers['data']

        answers = sanitize(answers)

        budget = map_2023.map_budget_entry(year, nren, answers)
        if budget:
            result[BudgetEntry].append(budget)

        funding = map_2023.map_fundingsource_entry(year, nren, answers)
        if funding:
            result[FundingSource].append(funding)

        charging = map_2023.map_charging_structure_entry(year, nren, answers)
        if charging:
            result[ChargingStructure].append(charging)

        nren_staff = map_2023.map_nren_staff(year, nren, answers)
        if nren_staff:
            result[NrenStaff].append(nren_staff)

        parent = map_2023.map_parent_orgs(nren, year, answers)
        if parent:
            result[ParentOrganization].append(parent)

        subs = map_2023.map_sub_orgs(nren, year, answers)
        for sub in subs:
            result[SubOrganization].append(sub)

        ec_projects = map_2023.map_ec_projects(nren, year, answers)
        for ec_project in ec_projects:
            result[ECProject].append(ec_project)

        policy = map_2023.map_policy(nren, year, answers)
        if policy:
            result[Policy].append(policy)

        traffic_volume = map_2023.map_traffic_volume(nren, year, answers)
        if traffic_volume:
            result[TrafficVolume].append(traffic_volume)

        institution_urls = map_2023.map_institution_urls(nren, year, answers)
        if institution_urls:
            result[InstitutionURLs].append(institution_urls)

        central_procurement = map_2023.map_central_procurement(nren, year, answers)
        if central_procurement:
            result[CentralProcurement].append(central_procurement)

        service_management = map_2023.map_service_management(nren, year, answers)
        if service_management:
            result[ServiceManagement].append(service_management)

        service_user_types = map_2023.map_service_user_types(nren, year, answers)
        for service_user_type in service_user_types:
            result[ServiceUserTypes].append(service_user_type)

        # https://jira.software.geant.org/browse/COMP-371
        # eosc_listings = map_2023.map_eosc_listings(nren, year, answers)
        # if eosc_listings:
        #     result[EOSCListings].append(eosc_listings)

        standards = map_2023.map_standards(nren, year, answers)
        if standards:
            result[Standards].append(standards)

        crisis_exercises = map_2023.map_crisis_exercises(nren, year, answers)
        if crisis_exercises:
            result[CrisisExercises].append(crisis_exercises)

        security_controls = map_2023.map_security_controls(nren, year, answers)
        if security_controls:
            result[SecurityControls].append(security_controls)

        connected_proportions = map_2023.map_connected_proportion(nren, year, answers)
        for connected_proportion in connected_proportions:
            result[ConnectedProportion].append(connected_proportion)

        connectivity_levels = map_2023.map_connectivity_levels(nren, year, answers)
        for connectivity_level in connectivity_levels:
            result[ConnectivityLevel].append(connectivity_level)

        connection_carriers = map_2023.map_connection_carriers(nren, year, answers)
        for connection_carrier in connection_carriers:
            result[ConnectionCarrier].append(connection_carrier)

        connectivity_loads = map_2023.map_connectivity_loads(nren, year, answers)
        for connectivity_load in connectivity_loads:
            result[ConnectivityLoad].append(connectivity_load)

        connectivity_growths = map_2023.map_connectivity_growth(nren, year, answers)
        for connectivity_growth in connectivity_growths:
            result[ConnectivityGrowth].append(connectivity_growth)

        commercial_connectivities = map_2023.map_commercial_connectivity(nren, year, answers)
        for commercial_connectivity in commercial_connectivities:
            result[CommercialConnectivity].append(commercial_connectivity)

        commercial_charging_levels = map_2023.map_commercial_charging_levels(nren, year, answers)
        for commercial_charging_level in commercial_charging_levels:
            result[CommercialChargingLevel].append(commercial_charging_level)

        remote_campuses = map_2023.map_remote_campuses(nren, year, answers)
        if remote_campuses:
            result[RemoteCampuses].append(remote_campuses)

        dark_fibre_lease = map_2023.map_dark_fibre_lease(nren, year, answers)
        if dark_fibre_lease:
            result[DarkFibreLease].append(dark_fibre_lease)

        dark_fibre_installed = map_2023.map_dark_fibre_installed(nren, year, answers)
        if dark_fibre_installed:
            result[DarkFibreInstalled].append(dark_fibre_installed)

        fibre_light = map_2023.map_fibre_light(nren, year, answers)
        if fibre_light:
            result[FibreLight].append(fibre_light)

        network_map_urls = map_2023.map_network_map_urls(nren, year, answers)
        if network_map_urls:
            result[NetworkMapUrls].append(network_map_urls)

        monitoring_tools = map_2023.map_monitoring_tools(nren, year, answers)
        if monitoring_tools:
            result[MonitoringTools].append(monitoring_tools)

        passive_monitoring = map_2023.map_passive_monitoring(nren, year, answers)
        if passive_monitoring:
            result[PassiveMonitoring].append(passive_monitoring)

        traffic_statistics = map_2023.map_traffic_statistics(nren, year, answers)
        if traffic_statistics:
            result[TrafficStatistics].append(traffic_statistics)

        siem_vendors = map_2023.map_siem_vendors(nren, year, answers)
        if siem_vendors:
            result[SiemVendors].append(siem_vendors)

        certificate_providers = map_2023.map_certificate_providers(nren, year, answers)
        if certificate_providers:
            result[CertificateProviders].append(certificate_providers)

        weathermap = map_2023.map_weathermap_entry(nren, year, answers)
        if weathermap:
            result[WeatherMap].append(weathermap)

        pert_team = map_2023.map_pert_team(nren, year, answers)
        if pert_team:
            result[PertTeam].append(pert_team)

        alienwave = map_2023.map_alienwave_entry(nren, year, answers)
        if alienwave:
            result[AlienWave].append(alienwave)

        capacity = map_2023.map_capacity_entry(nren, year, answers)
        if capacity:
            result[Capacity].append(capacity)

        external_connections = map_2023.map_external_connections(nren, year, answers)
        if external_connections:
            result[ExternalConnections].append(external_connections)

        non_re_peers = map_2023.map_non_re_peers(nren, year, answers)
        if non_re_peers:
            result[NonREPeers].append(non_re_peers)

        traffic_ratio = map_2023.map_traffic_ratio(nren, year, answers)
        if traffic_ratio:
            result[TrafficRatio].append(traffic_ratio)

        ops_automation = map_2023.map_ops_automation(nren, year, answers)
        if ops_automation:
            result[OpsAutomation].append(ops_automation)

        nfv = map_2023.map_network_function_virtualisation(nren, year, answers)
        if nfv:
            result[NetworkFunctionVirtualisation].append(nfv)

        network_automation = map_2023.map_network_automation(nren, year, answers)
        if network_automation:
            result[NetworkAutomation].append(network_automation)

        nren_services = map_2023.map_nren_services(nren, year, answers)
        for nren_service in nren_services:
            result[NRENService].append(nren_service)

    return result
