import logging
from typing import Any, TypedDict, List, Dict

from flask import Blueprint, request
from flask_login import login_required, current_user  # type: ignore
from sqlalchemy import delete, select
from sqlalchemy.orm import joinedload, load_only

from compendium_v2.db import db
from compendium_v2.db.auth_model import User
from compendium_v2.db.presentation_models import NREN, PreviewYear
from compendium_v2.db.survey_models import Survey, SurveyResponse, SurveyNotes, SurveyStatus, \
    RESPONSE_NOT_STARTED, RESPONSE_NOT_COMPLETED
from compendium_v2.publishers.survey_publisher import publish
from compendium_v2.routes import common
from compendium_v2.auth.session_management import admin_required


routes = Blueprint('survey', __name__)
logger = logging.getLogger(__name__)


LIST_SURVEYS_RESPONSE_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',
    'definitions': {
        'response': {
            'type': 'object',
            'properties': {
                'nren': {
                    'type': 'object',
                    'properties': {
                        'name': {'type': 'string'},
                        'id': {'type': 'integer'}
                    },
                },
                'status': {'type': 'string'},
                'lock_description': {'type': 'string'},
                'notes': {'type': 'string'},
                'valid': {'type': 'boolean'}
            },
            'required': ['nren', 'status', 'lock_description', 'valid'],
            'additionalProperties': False
        },
        'survey': {
            'type': 'object',
            'properties': {
                'year': {'type': 'integer'},
                'status': {'type': 'string'},
                'responses': {'type': 'array', 'items': {'$ref': '#/definitions/response'}},
            },
            'required': ['year', 'status', 'responses'],
            'additionalProperties': False
        }
    },
    'type': 'array',
    'items': {'$ref': '#/definitions/survey'}
}

SURVEY_ACTIVE_YEAR_RESPONSE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "properties": {
        "year": {
            "type": "string"
        }
    },
    "required": ["year"],
    "additionalProperties": False
}


@routes.route('/active/year', methods=['GET'])
@common.require_accepts_json
@login_required
def get_active_survey_year() -> Any:
    """
    retrieve a year of latest active survey

    response will be formatted as:

    .. asjson::
        compendium_v2.routes.survey.SURVEY_ACTIVE_YEAR_RESPONSE_SCHEMA

    """

    survey_record = db.session.query(Survey.year).filter(
        Survey.status == SurveyStatus.open
    ).order_by(Survey.year.desc()).first()
    if survey_record:
        year = survey_record.year
        return {'year': year}
    else:
        return {'message': 'No open survey found.'}, 404


@routes.route('/list', methods=['GET'])
@common.require_accepts_json
@login_required
def list_surveys() -> Any:
    """
    retrieve a list of surveys and responses, including their status

    response will be formatted as:

    .. asjson::
        compendium_v2.routes.survey.LIST_SURVEYS_RESPONSE_SCHEMA

    """

    if not (current_user.is_admin or current_user.is_observer):
        return {'message': 'Insufficient privileges to access this resource'}, 403

    loader = joinedload(Survey.responses).load_only(SurveyResponse.status, SurveyResponse.valid)
    nren_loader = loader.joinedload(SurveyResponse.nren).load_only(NREN.name)
    notes_loader = loader.joinedload(SurveyResponse.notes).load_only(SurveyNotes.notes)
    user_loader = loader.joinedload(SurveyResponse.locked_by_user).load_only(User.fullname, User.email)

    surveys = db.session.scalars(
        select(Survey).options(
            load_only(Survey.year, Survey.status),
            loader,
            nren_loader,
            notes_loader,
            user_loader
        ).order_by(Survey.year.desc())
    ).unique()

    class SurveyDict(TypedDict):
        year: int
        status: str
        responses: List[Dict[str, Any]]

    def _get_response(response: SurveyResponse) -> Dict[str, Any]:
        res = {
            "nren": {
                'name': response.nren.name,
                'id': response.nren.id
            },
            "status": response.status.value,
            "lock_description": response.lock_description,
            "notes": response.notes.notes if response.notes else "",
            "valid": response.valid
        }
        return res

    def sort_key(response: SurveyResponse):
        return response.status.value + response.nren.name.lower()

    all_nrens = db.session.scalars(select(NREN)).all()
    nrens = {nren.name: nren for nren in all_nrens}
    nren_names = set(nrens.keys())

    all_surveys: List[SurveyDict] = []

    for survey in surveys:
        status = survey.status
        responses = [_get_response(r) for r in sorted(survey.responses, key=sort_key)]
        survey_dict: SurveyDict = {
            "year": survey.year,
            "status": status.value,
            "responses": responses
        }

        if status in [SurveyStatus.open, SurveyStatus.closed, SurveyStatus.published]:
            # add in nrens without a response and set a custom response status depending on the survey status
            nrens_with_responses = set([r["nren"]["name"] for r in survey_dict["responses"]])
            missing_responses = nren_names.difference(nrens_with_responses)
            for nren_name in sorted(missing_responses, key=str.lower):
                nren = nrens[nren_name]
                if not nren.can_answer_survey():
                    # exclude nrens that are not allowed to answer the survey
                    continue
                responses.append({
                    "status": RESPONSE_NOT_STARTED if status == SurveyStatus.open else RESPONSE_NOT_COMPLETED,
                    "lock_description": "",
                    "valid": False,
                    "nren": {
                        'name': nren_name,
                        'id': nren.id
                    },
                })

        all_surveys.append(survey_dict)

    return all_surveys


@routes.route('/new', methods=['POST'])
@common.require_accepts_json
@admin_required
def start_new_survey() -> Any:
    """
    endpoint to initiate a new survey

    :returns: ``{'success': True}`` or a 400 or 404 status with a descriptive message
    """
    all_surveys = db.session.scalars(select(Survey).options(load_only(Survey.status)))
    if any([survey.status != SurveyStatus.published for survey in all_surveys]):
        return {'message': 'All earlier surveys should be published before starting a new one'}, 400

    last_survey = db.session.scalar(
        select(Survey).order_by(Survey.year.desc()).limit(1)
    )

    if not last_survey:
        return {'message': 'No surveys found'}, 404

    new_year = last_survey.year + 1
    new_survey = last_survey.survey
    new_survey = Survey(year=new_year, survey=new_survey, status=SurveyStatus.closed)
    db.session.add(new_survey)
    db.session.commit()

    return {'success': True}


@routes.route('/open/<int:year>', methods=['POST'])
@common.require_accepts_json
@admin_required
def open_survey(year) -> Any:
    """
    endpoint to open a survey to the nrens

    :returns: ``{'success': True}`` or a 400 or 404 status with a descriptive message
    """
    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if not survey:
        return {'message': 'Survey not found'}, 404

    if survey.status != SurveyStatus.closed:
        return {'message': 'Survey is not closed and can therefore not be opened'}, 400

    all_surveys = db.session.scalars(select(Survey))
    if any([s.status == SurveyStatus.open for s in all_surveys]):
        return {'message': 'There already is an open survey'}, 400

    survey.status = SurveyStatus.open
    db.session.commit()

    return {'success': True}


@routes.route('/close/<int:year>', methods=['POST'])
@common.require_accepts_json
@admin_required
def close_survey(year) -> Any:
    """
    endpoint to close a survey to the nrens

    :returns: ``{'success': True}`` or a 400 or 404 status with a descriptive message
    """
    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if not survey:
        return {'message': 'Survey not found'}, 404

    if survey.status != SurveyStatus.open:
        return {'message': 'Survey is not open and can therefore not be closed'}, 400

    survey.status = SurveyStatus.closed
    db.session.commit()

    return {'success': True}


@routes.route('/preview/<int:year>', methods=['POST'])
@common.require_accepts_json
@admin_required
def preview_survey(year) -> Any:
    """
    endpoint to preview a survey on the compendium website

    :returns: ``{'success': True}`` or a 400 or 404 status with a descriptive message
    """
    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if not survey:
        return {'message': 'Survey not found'}, 404

    if survey.status not in [SurveyStatus.closed, SurveyStatus.preview]:
        return {'message': 'Survey is not closed or in preview and can therefore not be published for preview'}, 400

    try:
        publish(year, dry_run=False)
    except ValueError as e:
        return {'message': str(e)}, 400

    preview = db.session.scalar(select(PreviewYear).where(PreviewYear.year == year))
    if not preview:
        db.session.add(PreviewYear(year=year))

    survey.status = SurveyStatus.preview
    db.session.commit()

    return {'success': True}


@routes.route('/publish/<int:year>', methods=['POST'])
@common.require_accepts_json
@admin_required
def publish_survey(year) -> Any:
    """
    endpoint to publish a survey to the compendium website

    :returns: ``{'success': True}`` or a 400 or 404 status with a descriptive message
    """
    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if not survey:
        return {'message': 'Survey not found'}, 404

    if survey.status not in [SurveyStatus.preview, SurveyStatus.published]:
        return {'message': 'Survey is not in preview or published and can therefore not be published'}, 400

    dry_run = request.args.get('dry_run', False)
    dry_run = bool(dry_run)
    try:
        output = publish(year, dry_run=dry_run)
    except ValueError as e:
        return {'message': str(e)}, 400

    if not dry_run:
        db.session.execute(delete(PreviewYear).where(PreviewYear.year == year))

        survey.status = SurveyStatus.published
        db.session.commit()

    result = {'success': True}
    if output:
        result['message'] = output

    return result


@routes.route('/<int:year>/<int:nren_id>/notes', methods=['POST'])
@common.require_accepts_json
@admin_required
def survey_notes(year, nren_id) -> Any:
    """
    endpoint to add management notes to a survey response

    :returns: ``{'success': True}`` or a 400 or 404 status with a descriptive message
    """

    response = db.session.scalar(
        select(SurveyResponse).where(SurveyResponse.survey_year == year, SurveyResponse.nren_id == nren_id)
    )
    if not response:
        return {'message': 'Survey response not found'}, 404

    body = request.get_json()
    if 'notes' not in body:
        return {'message': 'Notes not provided'}, 400

    existing_notes = response.notes

    value = body['notes'] or ''

    if existing_notes is None:
        notes = SurveyNotes(nren_id=nren_id, survey_year=year, survey=response, notes=value)
        response.notes = notes
        db.session.add(notes)
    else:
        existing_notes.notes = value

    db.session.commit()

    return {'success': True, 'notes': response.notes.notes}
