"""
test_data_dump
=========================

This script loads data from the local DB and dumps it in a format usable by various tests.
"""
import logging
import click
import pathlib
import json
from decimal import Decimal
from enum import Enum

from itertools import chain
from sqlalchemy import select
from typing import Type, Sequence, TypeVar

import compendium_v2
from compendium_v2.db import db
from compendium_v2.db.presentation_models import NREN, PresentationModel, Service
from compendium_v2.config import load
from compendium_v2.environment import setup_logging
from compendium_v2.routes.api import models
setup_logging()

logger = logging.getLogger('test_data_dump')


T = TypeVar('T', bound=PresentationModel)


def _get_dict_data(model: Type[T]) -> dict:
    data = model.__dict__.copy()
    del data['_sa_instance_state']
    for key, val in list(data.items()):
        if isinstance(val, Enum):
            data[key] = val.value
        else:
            if not isinstance(val, str) and not isinstance(val, int):
                print(type(val))
    return data


def extract_model_data(model: Type[T]):
    data = model.__dict__.copy()
    del data['_sa_instance_state']
    for key, val in list(data.items()):
        if isinstance(val, list):
            if val and isinstance(val[0], str):
                continue
            if all(isinstance(v, dict) for v in val):
                continue
            data[key] = [_get_dict_data(v) for v in val]
        elif isinstance(val, db.Model):
            del data[key]
        elif isinstance(val, Enum):
            data[key] = val.value
        elif isinstance(val, Decimal):
            data[key] = float(val)
        else:
            pass
    return data


def get_data(table_class: Type[T]) -> Sequence[T]:
    if table_class == NREN:
        data = [extract_model_data(n) for n in db.session.scalars(select(NREN).order_by(NREN.name.asc()))]
        return list(chain(data))
    select_statement = select(table_class).join(NREN).order_by(NREN.name.asc(), table_class.year.asc())

    data = [extract_model_data(m) for m in db.session.scalars(select_statement)]
    return list(data)


def _cli(app):
    TEST_DATA_DIR = pathlib.Path(__file__).parent.parent.parent / 'test' / 'data' / 'presentation_models_data'
    TEST_DATA_DIR.mkdir(parents=True, exist_ok=True)

    logger.info(f'Inserting data into {TEST_DATA_DIR}')
    with app.app_context():
        for model, url, schema in models:
            logger.info(f'Dumping data for {model.__name__}')
            with open(TEST_DATA_DIR / f'{model.__name__}.json', 'w') as f:
                data = get_data(model)
                f.write(json.dumps(data, indent=2))
        # dump services as well
        with open(TEST_DATA_DIR / 'Service.json', 'w') as f:
            data = [extract_model_data(s) for s in db.session.scalars(select(Service).order_by(Service.name_key.asc()))]
            f.write(json.dumps(data, indent=2))
    logger.info('Done')


@click.command()
@click.option('--config', type=click.STRING, default='config.json')
def cli(config):
    app_config = load(open(config, 'r'))
    app = compendium_v2._create_app_with_db(app_config)
    _cli(app)


if __name__ == "__main__":
    cli()
